#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

from tempest.lib.common.utils import data_utils
from tempest.lib import decorators
from tempest.lib import exceptions as lib_exc

from ironic_tempest_plugin.tests.api.admin import api_microversion_fixture
from ironic_tempest_plugin.tests.api import base


EXAMPLE_STEPS = [{
    'interface': 'bios',
    'step': 'apply_configuration',
    'args': {},
    'priority': 10
}]


def _get_random_trait():
    return data_utils.rand_name('CUSTOM', '').replace('-', '_')


class TestDeployTemplates(base.BaseBaremetalTest):
    """Tests for deploy templates."""

    min_microversion = '1.55'

    def setUp(self):
        super(TestDeployTemplates, self).setUp()
        self.name = _get_random_trait()
        self.steps = copy.deepcopy(EXAMPLE_STEPS)
        _, self.template = self.create_deploy_template(self.name,
                                                       steps=self.steps)

    @decorators.idempotent_id('8f39c794-4e6f-42fc-bfcb-e1f6eafff2e9')
    def test_create_deploy_template_specifying_uuid(self):
        name = _get_random_trait()
        uuid = data_utils.rand_uuid()

        _, template = self.create_deploy_template(name=name, steps=self.steps,
                                                  uuid=uuid)

        _, body = self.client.show_deploy_template(uuid)
        self._assertExpected(template, body)

    @decorators.idempotent_id('7ac8fdd5-928b-4fbc-8341-3ebaad9def5e')
    def test_delete_deploy_template(self):
        self.delete_deploy_template(self.template['uuid'])

        self.assertRaises(lib_exc.NotFound, self.client.show_deploy_template,
                          self.template['uuid'])

    @decorators.idempotent_id('f424dd67-46c3-4169-b8f0-7e0c18b70437')
    def test_show_deploy_template(self):
        _, template = self.client.show_deploy_template(self.template['uuid'])
        self._assertExpected(self.template, template)
        self.assertEqual(self.name, template['name'])
        self.assertEqual(self.steps, template['steps'])
        self.assertIn('uuid', template)
        self.assertEqual({}, template['extra'])

    @decorators.idempotent_id('2fd98e9a-10ce-405a-a32c-0d6079766183')
    def test_show_deploy_template_with_links(self):
        _, template = self.client.show_deploy_template(self.template['uuid'])
        self.assertIn('links', template)
        self.assertEqual(2, len(template['links']))
        self.assertIn(template['uuid'], template['links'][0]['href'])

    @decorators.idempotent_id('cec2a01d-07af-4062-a8b0-9a1703f65bcf')
    def test_list_deploy_templates(self):
        _, body = self.client.list_deploy_templates()
        self.assertIn(self.template['uuid'],
                      [i['uuid'] for i in body['deploy_templates']])
        # Verify self links.
        for template in body['deploy_templates']:
            self.validate_self_link('deploy_templates', template['uuid'],
                                    template['links'][0]['href'])

    @decorators.idempotent_id('89aea2bf-c094-445f-b869-9fd56d1dfe5a')
    def test_list_with_limit(self):
        for i in range(2):
            name = _get_random_trait()
            self.create_deploy_template(name, steps=self.steps)

        _, body = self.client.list_deploy_templates(limit=3)

        next_marker = body['deploy_templates'][-1]['uuid']
        self.assertIn(next_marker, body['next'])

    @decorators.idempotent_id('c09c917e-c4d2-4148-9b6a-459bd126ed7c')
    def test_list_deploy_templates_detail(self):
        uuids = [
            self.create_deploy_template(_get_random_trait(), steps=self.steps)
            [1]['uuid'] for i in range(0, 5)]

        _, body = self.client.list_deploy_templates(detail=True)

        templates_dict = dict((template['uuid'], template)
                              for template in body['deploy_templates']
                              if template['uuid'] in uuids)

        for uuid in uuids:
            self.assertIn(uuid, templates_dict)
            template = templates_dict[uuid]
            self.assertIn('name', template)
            self.assertEqual(self.steps, template['steps'])
            self.assertIn('uuid', template)
            self.assertEqual({}, template['extra'])
            # Verify self link.
            self.validate_self_link('deploy_templates', template['uuid'],
                                    template['links'][0]['href'])

    @decorators.idempotent_id('a6cf1ade-e19a-41e2-b151-13ecf0d8f08c')
    def test_update_deploy_template_replace(self):
        new_name = _get_random_trait()
        new_steps = [{
            'interface': 'raid',
            'step': 'create_configuration',
            'args': {},
            'priority': 10,
        }]

        patch = [{'path': '/name', 'op': 'replace', 'value': new_name},
                 {'path': '/steps', 'op': 'replace', 'value': new_steps}]

        self.client.update_deploy_template(self.template['uuid'], patch)

        _, body = self.client.show_deploy_template(self.template['uuid'])
        self.assertEqual(new_name, body['name'])
        self.assertEqual(new_steps, body['steps'])

    @decorators.idempotent_id('bb168c63-452b-4065-9a81-77853ca9540a')
    def test_update_deploy_template_add(self):
        new_steps = [
            {
                'interface': 'bios',
                'step': 'cache_bios_settings',
                'args': {},
                'priority': 20
            },
            {
                'interface': 'bios',
                'step': 'factory_reset',
                'args': {},
                'priority': 30
            },
        ]

        patch = [{'path': '/steps/1', 'op': 'add', 'value': new_steps[0]},
                 {'path': '/steps/2', 'op': 'add', 'value': new_steps[1]}]

        self.client.update_deploy_template(self.template['uuid'], patch)

        _, body = self.client.show_deploy_template(self.template['uuid'])
        self.assertEqual(self.steps + new_steps, body['steps'])

    @decorators.idempotent_id('2aa204a2-1d50-48fd-8b76-d2ed15586d50')
    def test_update_deploy_template_mixed_ops(self):
        new_name = _get_random_trait()
        new_steps = [
            {
                'interface': 'bios',
                'step': 'apply_configuration',
                'args': {},
                'priority': 20
            },
            {
                'interface': 'bios',
                'step': 'apply_configuration',
                'args': {},
                'priority': 30
            },
        ]

        patch = [{'path': '/name', 'op': 'replace', 'value': new_name},
                 {'path': '/steps/0', 'op': 'replace', 'value': new_steps[0]},
                 {'path': '/steps/0', 'op': 'remove'},
                 {'path': '/steps/0', 'op': 'add', 'value': new_steps[1]}]

        self.client.update_deploy_template(self.template['uuid'], patch)

        _, body = self.client.show_deploy_template(self.template['uuid'])
        self.assertEqual(new_name, body['name'])
        self.assertEqual([new_steps[1]], body['steps'])


class TestDeployTemplatesWithJsonExtSupport(base.BaseBaremetalTest):
    """Tests for deploy templates to validate appending .json extension to a

    template's `name` or `uuid` in API versions 1.90 and prior works.
    """

    max_microversion = '1.90'  # Max API version allowed for testing: 1.90
    min_microversion = '1.90'  # Min API version allowed for testing: 1.90

    def setUp(self):
        super(TestDeployTemplatesWithJsonExtSupport, self).setUp()
        self.name = _get_random_trait()
        self.steps = copy.deepcopy(EXAMPLE_STEPS)
        _, self.template = self.create_deploy_template(self.name,
                                                       steps=self.steps)

    @decorators.idempotent_id('df2fabbd-3c8b-4979-8f13-c58c96122adb')
    def test_delete_deploy_template_with_json(self):
        """Delete a template while appending .json extension to its uuid"""
        self.delete_deploy_template('%s.json' % self.template['uuid'])

        self.assertRaises(lib_exc.NotFound, self.client.show_deploy_template,
                          self.template['uuid'])

    @decorators.idempotent_id('8b64a54d-ead7-4b3f-8fae-a1b8bdefe3ee')
    def test_show_deploy_template_with_json(self):
        """Show a template while appending .json extension to its uuid"""
        _, template = self.client.show_deploy_template('%s.json' %
                                                       self.template['uuid'])
        self._assertExpected(self.template, template)
        self.assertEqual(self.name, template['name'])
        self.assertEqual(self.steps, template['steps'])
        self.assertIn('uuid', template)
        self.assertEqual({}, template['extra'])


class TestDeployTemplatesWithoutJsonExtSupport(base.BaseBaremetalTest):
    """Tests for deploy templates to validate appending .json extension to a

    template's `name` or `uuid` in API versions later than 1.90 is disabled.
    """

    min_microversion = '1.91'  # Min API version allowed for testing: 1.91

    def setUp(self):
        super(TestDeployTemplatesWithoutJsonExtSupport, self).setUp()
        self.name = _get_random_trait()
        self.steps = copy.deepcopy(EXAMPLE_STEPS)
        _, self.template = self.create_deploy_template(self.name,
                                                       steps=self.steps)

    @decorators.idempotent_id('61d05189-9f84-4973-835b-a860b655dfe3')
    def test_delete_deploy_template_with_json(self):
        """Trying to delete a template while appending .json ext 404s"""
        self.assertRaises(lib_exc.NotFound, self.client.delete_deploy_template,
                          '%s.json' % self.template['uuid'])

    @decorators.idempotent_id('432429dd-b964-4133-aa46-cd43a7cd9c37')
    def test_show_deploy_template_with_json(self):
        """Trying to show a template while appending .json ext 404s"""
        self.assertRaises(lib_exc.NotFound, self.client.show_deploy_template,
                          '%s.json' % self.template['uuid'])


class TestDeployTemplatesOldAPI(base.BaseBaremetalTest):
    """Negative tests for deploy templates using an old API version."""

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('e9481a0d-23e0-4757-bc11-c3c9ab9d3839')
    def test_create_deploy_template_old_api(self):
        # With deploy templates support, ironic returns 404. Without, 405.
        self.assertRaises((lib_exc.NotFound, lib_exc.UnexpectedResponseCode),
                          self.create_deploy_template,
                          name=_get_random_trait(), steps=EXAMPLE_STEPS)

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('0d3af2aa-ba53-4c8a-92d4-91f9b4179fe7')
    def test_update_deploy_template_old_api(self):
        patch = [{'path': '/name', 'op': 'replace',
                  'value': _get_random_trait()}]

        # With deploy templates support, ironic returns 404. Without, 405.
        self.assertRaises((lib_exc.NotFound, lib_exc.UnexpectedResponseCode),
                          self.client.update_deploy_template,
                          _get_random_trait(), patch)

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('1646b1e5-ab81-45a8-9ea0-30444a4dcaa2')
    def test_delete_deploy_template_old_api(self):
        # With deploy templates support, ironic returns 404. Without, 405.
        self.assertRaises((lib_exc.NotFound, lib_exc.UnexpectedResponseCode),
                          self.client.delete_deploy_template,
                          _get_random_trait())

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('819480ac-f36a-4402-b1d5-504d7cf55b1f')
    def test_list_deploy_templates_old_api(self):
        self.assertRaises(lib_exc.NotFound,
                          self.client.list_deploy_templates)


class TestDeployTemplatesNegative(base.BaseBaremetalTest):
    """Negative tests for deploy templates."""

    min_microversion = '1.55'

    def setUp(self):
        super(TestDeployTemplatesNegative, self).setUp()
        self.useFixture(
            api_microversion_fixture.APIMicroversionFixture(
                self.min_microversion)
        )
        self.steps = EXAMPLE_STEPS

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('a4085c08-e718-4c2f-a796-0e115b659243')
    def test_create_deploy_template_invalid_name(self):
        name = 'invalid-name'
        self.assertRaises(lib_exc.BadRequest,
                          self.create_deploy_template, name=name,
                          steps=self.steps)

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('6390acc4-9490-4b23-8b4c-41888a78c9b7')
    def test_create_deploy_template_duplicated_deploy_template_name(self):
        name = _get_random_trait()
        self.create_deploy_template(name=name, steps=self.steps)
        self.assertRaises(lib_exc.Conflict, self.create_deploy_template,
                          name=name, steps=self.steps)

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('ed3f0cec-13e8-4175-9fdb-d129e7b7fe10')
    def test_create_deploy_template_no_mandatory_field_name(self):
        self.assertRaises(lib_exc.BadRequest, self.create_deploy_template,
                          name=None, steps=self.steps)

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('af5dd0df-d903-463f-9535-9e4e9d6fd576')
    def test_create_deploy_template_no_mandatory_field_steps(self):
        self.assertRaises(lib_exc.BadRequest, self.create_deploy_template,
                          name=_get_random_trait())

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('cbd33bc5-7602-40b7-943e-3e92217567a3')
    def test_create_deploy_template_malformed_steps(self):
        steps = {'key': 'value'}
        self.assertRaises(lib_exc.BadRequest, self.create_deploy_template,
                          name=_get_random_trait(), steps=steps)

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('2a562fca-f377-4a6e-b332-37ee82d3a983')
    def test_create_deploy_template_malformed_deploy_template_uuid(self):
        uuid = 'malformed:uuid'
        self.assertRaises(lib_exc.BadRequest, self.create_deploy_template,
                          name=_get_random_trait(), steps=self.steps,
                          uuid=uuid)

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('2c006994-88ca-43b7-b605-897d479229d9')
    def test_show_deploy_template_nonexistent(self):
        self.assertRaises(lib_exc.NotFound, self.client.show_deploy_template,
                          data_utils.rand_uuid())

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('5a815f37-f015-4d68-9b22-099504f74805')
    def test_update_deploy_template_remove_mandatory_field_steps(self):
        name = _get_random_trait()
        _, template = self.create_deploy_template(name=name, steps=self.steps)

        # Removing one item from the collection
        self.assertRaises(lib_exc.BadRequest,
                          self.client.update_deploy_template,
                          template['uuid'],
                          [{'path': '/steps/0', 'op': 'remove'}])

        # Removing the collection
        self.assertRaises(lib_exc.BadRequest,
                          self.client.update_deploy_template,
                          template['uuid'],
                          [{'path': '/steps', 'op': 'remove'}])

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('ee852ebb-a601-4593-9d59-063fcbc8f964')
    def test_update_deploy_template_remove_mandatory_field_name(self):
        name = _get_random_trait()
        _, template = self.create_deploy_template(name=name, steps=self.steps)
        self.assertRaises(lib_exc.BadRequest,
                          self.client.update_deploy_template,
                          template['uuid'],
                          [{'path': '/name', 'op': 'remove'}])

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('e59bf38d-272f-4490-b21e-9db217f11378')
    def test_update_deploy_template_replace_empty_name(self):
        name = _get_random_trait()
        _, template = self.create_deploy_template(name=name, steps=self.steps)
        self.assertRaises(lib_exc.BadRequest,
                          self.client.update_deploy_template,
                          template['uuid'],
                          [{'path': '/name', 'op': 'replace', 'value': ''}])



#
# Copyright 2017 Mirantis Inc.
#
# Copyright (c) 2022 Dell Inc. or its subsidiaries.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from tempest.common import utils
from tempest import config
from tempest.lib import decorators

from ironic_tempest_plugin.tests.scenario import \
    baremetal_standalone_manager as bsm

CONF = config.CONF


class BaremetalDriverIscsiWholedisk(bsm.BaremetalStandaloneScenarioTest):

    api_microversion = '1.31'  # to set the deploy_interface
    if 'redfish' in CONF.baremetal.enabled_hardware_types:
        driver = 'redfish'
    else:
        driver = 'ipmi'

    deploy_interface = 'iscsi'
    image_ref = CONF.baremetal.whole_disk_image_ref
    wholedisk_image = True

    @classmethod
    def skip_checks(cls):
        super(BaremetalDriverIscsiWholedisk, cls).skip_checks()
        if cls.driver == 'redfish':
            skip_msg = ("Test covered when using ipmi")
            raise cls.skipException(skip_msg)

    @decorators.idempotent_id('f25b71df-2150-45d7-a780-7f5b07124808')
    @utils.services('image', 'network')
    def test_ip_access_to_server(self):
        self.boot_and_verify_node()


class BaremetalDriverDirectWholedisk(bsm.BaremetalStandaloneScenarioTest):

    api_microversion = '1.31'  # to set the deploy_interface
    if 'redfish' in CONF.baremetal.enabled_hardware_types:
        driver = 'redfish'
    else:
        driver = 'ipmi'
    deploy_interface = 'direct'
    image_ref = CONF.baremetal.whole_disk_image_ref
    wholedisk_image = True

    @classmethod
    def skip_checks(cls):
        super(BaremetalDriverDirectWholedisk, cls).skip_checks()
        if cls.driver == 'ipmi':
            skip_msg = ("Test covered when using redfish")
            raise cls.skipException(skip_msg)

    @decorators.idempotent_id('c2db24e7-07dc-4a20-8f93-d4efae2bfd4e')
    @utils.services('image', 'network')
    def test_ip_access_to_server(self):
        self.boot_and_verify_node()


class BaremetalDriverIscsiPartitioned(bsm.BaremetalStandaloneScenarioTest):

    api_microversion = '1.31'  # to set the deploy_interface
    if 'redfish' in CONF.baremetal.enabled_hardware_types:
        driver = 'redfish'
    else:
        driver = 'ipmi'
    deploy_interface = 'iscsi'
    image_ref = CONF.baremetal.partition_image_ref
    wholedisk_image = False
    boot_option = 'netboot' if CONF.baremetal.partition_netboot else 'local'

    @classmethod
    def skip_checks(cls):
        super(BaremetalDriverIscsiPartitioned, cls).skip_checks()
        if cls.driver == 'ipmi':
            skip_msg = ("Test covered when using redfish")
            raise cls.skipException(skip_msg)

    @decorators.idempotent_id('7d0b205e-edbc-4e2d-9f6d-95cd74eefecb')
    @utils.services('image', 'network')
    def test_ip_access_to_server(self):
        self.boot_and_verify_node()


class BaremetalDriverDirectPartitioned(bsm.BaremetalStandaloneScenarioTest):

    api_microversion = '1.31'  # to set the deploy_interface
    if 'redfish' in CONF.baremetal.enabled_hardware_types:
        driver = 'redfish'
    else:
        driver = 'ipmi'
    deploy_interface = 'direct'
    image_ref = CONF.baremetal.partition_image_ref
    wholedisk_image = False
    boot_option = 'netboot' if CONF.baremetal.partition_netboot else 'local'

    @classmethod
    def skip_checks(cls):
        super(BaremetalDriverDirectPartitioned, cls).skip_checks()
        if cls.driver == 'redfish':
            skip_msg = ("Test covered when using ipmi")
            raise cls.skipException(skip_msg)

    @decorators.idempotent_id('7b4b2dcd-2bbb-44f5-991f-0964300af6b7')
    @utils.services('image', 'network')
    def test_ip_access_to_server(self):
        self.boot_and_verify_node()


class BaremetalDriverAnsibleWholedisk(bsm.BaremetalStandaloneScenarioTest):

    api_microversion = '1.31'  # to set the deploy_interface
    if 'redfish' in CONF.baremetal.enabled_hardware_types:
        driver = 'redfish'
    else:
        driver = 'ipmi'
    deploy_interface = 'ansible'
    image_ref = CONF.baremetal.whole_disk_image_ref
    wholedisk_image = True

    @decorators.unstable_test(bug='2101021')
    @decorators.idempotent_id('cde532cc-81ba-4489-b374-b4a85cc203eb')
    @utils.services('image', 'network')
    def test_ip_access_to_server(self):
        self.boot_and_verify_node()


class BaremetalIpmiRescueWholedisk(bsm.BaremetalStandaloneScenarioTest):

    api_microversion = '1.38'
    min_microversion = '1.38'
    driver = 'ipmi'
    rescue_interface = 'agent'
    image_ref = CONF.baremetal.whole_disk_image_ref
    wholedisk_image = True

    # NOTE(tiendc) Using direct deploy interface and a whole disk
    # image may lead to the bug:
    # https://bugs.launchpad.net/ironic/+bug/1750958
    # This is a workaround by using iscsi deploy interface.
    deploy_interface = 'iscsi'

    @decorators.idempotent_id('d6a1780f-c4bb-4136-8144-29e822e14d66')
    @utils.services('image', 'network')
    def test_rescue_mode(self):
        self.set_node_to_active(self.image_ref)
        self.rescue_unrescue()


class BaremetalIpmiRescuePartitioned(bsm.BaremetalStandaloneScenarioTest):

    api_microversion = '1.38'
    min_microversion = '1.38'
    driver = 'ipmi'
    rescue_interface = 'agent'
    image_ref = CONF.baremetal.partition_image_ref
    wholedisk_image = False
    boot_option = 'netboot' if CONF.baremetal.partition_netboot else 'local'

    # NOTE(jroll) the ansible deploy interface doesn't support partition images
    # with netboot mode. Since that's what is happening here, explicitly choose
    # a deploy interface to be sure we don't end up with a node using the
    # ansible interface here.
    deploy_interface = 'iscsi'

    @decorators.idempotent_id('113acd0a-9872-4631-b3ee-54da7e3bb262')
    @utils.services('image', 'network')
    def test_rescue_mode(self):
        self.set_node_to_active(self.image_ref)
        self.rescue_unrescue()


class BaremetalIloDirectWholediskHttpLink(
        bsm.BaremetalStandaloneScenarioTest):

    api_microversion = '1.31'  # to set the deploy_interface
    driver = 'ilo'
    deploy_interface = 'direct'
    boot_interface = 'ilo-virtual-media'
    image_ref = CONF.baremetal.whole_disk_image_url
    image_checksum = CONF.baremetal.whole_disk_image_checksum
    wholedisk_image = True

    @decorators.idempotent_id('c2db24e7-b9bb-44df-6765-e60134346fd0')
    @utils.services('network')
    def test_ip_access_to_server(self):
        self.boot_and_verify_node()


class BaremetalIloDirectPartitioned(bsm.BaremetalStandaloneScenarioTest):

    api_microversion = '1.31'  # to set the deploy_interface
    driver = 'ilo'
    deploy_interface = 'direct'
    boot_interface = 'ilo-virtual-media'
    image_ref = CONF.baremetal.partition_image_ref
    wholedisk_image = False

    @decorators.idempotent_id('ea85e19c-d8dc-4577-4d05-fbbbe4ce3f8c')
    @utils.services('image', 'network')
    def test_ip_access_to_server(self):
        self.boot_and_verify_node()


class BaremetalIloIscsiWholediskHttpLink(
        bsm.BaremetalStandaloneScenarioTest):

    api_microversion = '1.31'  # to set the deploy_interface
    driver = 'ilo'
    deploy_interface = 'iscsi'
    boot_interface = 'ilo-virtual-media'
    image_ref = CONF.baremetal.whole_disk_image_url
    image_checksum = CONF.baremetal.whole_disk_image_checksum
    wholedisk_image = True

    @decorators.idempotent_id('71ccf06f-45db-8f93-afd0-d4efae2bfd4e')
    @utils.services('network')
    def test_ip_access_to_server(self):
        self.boot_and_verify_node()


class BaremetalIloIscsiPartitioned(bsm.BaremetalStandaloneScenarioTest):

    api_microversion = '1.31'  # to set the deploy_interface
    driver = 'ilo'
    deploy_interface = 'iscsi'
    boot_interface = 'ilo-virtual-media'
    image_ref = CONF.baremetal.partition_image_ref
    wholedisk_image = False

    @decorators.idempotent_id('d926c683-4d05-8252-b9bb-35babbed6e86')
    @utils.services('image', 'network')
    def test_ip_access_to_server(self):
        self.boot_and_verify_node()


class BaremetalIloPxeWholediskHttpLink(
        bsm.BaremetalStandaloneScenarioTest):

    api_microversion = '1.31'  # to set the deploy_interface
    driver = 'ilo'
    deploy_interface = 'direct'
    boot_interface = 'ilo-pxe'
    image_ref = CONF.baremetal.whole_disk_image_url
    image_checksum = CONF.baremetal.whole_disk_image_checksum
    wholedisk_image = True

    @decorators.idempotent_id('d926c683-1a32-edbc-07dc-95cd74eefecb')
    @utils.services('network')
    def test_ip_access_to_server(self):
        self.boot_and_verify_node()


class BaremetalIloPxePartitioned(bsm.BaremetalStandaloneScenarioTest):

    api_microversion = '1.31'  # to set the deploy_interface
    driver = 'ilo'
    deploy_interface = 'direct'
    boot_interface = 'ilo-pxe'
    image_ref = CONF.baremetal.partition_image_ref
    wholedisk_image = False

    @decorators.idempotent_id('71ccf06f-07dc-4577-6869-1b1bfa423b9b')
    @utils.services('image', 'network')
    def test_ip_access_to_server(self):
        self.boot_and_verify_node()


class BaremetalIloIPxeWholediskHttpLink(
        bsm.BaremetalStandaloneScenarioTest):

    api_microversion = '1.31'  # to set the deploy_interface
    driver = 'ilo'
    deploy_interface = 'direct'
    boot_interface = 'ilo-ipxe'
    image_ref = CONF.baremetal.whole_disk_image_url
    image_checksum = CONF.baremetal.whole_disk_image_checksum
    wholedisk_image = True

    @decorators.idempotent_id('d926c683-1a32-edbc-07dc-95cd74eefecb')
    @utils.services('network')
    def test_ip_access_to_server(self):
        self.boot_and_verify_node()


class BaremetalIlo5UefiHTTPSWholediskHttpsLink(
        bsm.BaremetalStandaloneScenarioTest):

    api_microversion = '1.31'  # to set the deploy_interface
    driver = 'ilo5'
    deploy_interface = 'direct'
    boot_interface = 'ilo-uefi-https'
    image_ref = CONF.baremetal.whole_disk_image_url
    image_checksum = CONF.baremetal.whole_disk_image_checksum
    wholedisk_image = True

    @decorators.idempotent_id('d926c683-1a32-edbc-07dc-95cd74eefecb')
    @utils.services('network')
    def test_ip_access_to_server(self):
        self.boot_and_verify_node()


class BaremetalRedfishDirectWholediskHttpLink(
        bsm.BaremetalStandaloneScenarioTest):

    api_microversion = '1.31'  # to set the deploy_interface
    driver = 'redfish'
    deploy_interface = 'direct'
    boot_interface = 'redfish-virtual-media'
    image_ref = CONF.baremetal.whole_disk_image_url
    image_checksum = CONF.baremetal.whole_disk_image_checksum
    wholedisk_image = True

    @decorators.idempotent_id('113acd0a-9872-4631-b3ee-54da7e3bb262')
    @utils.services('network')
    def test_ip_access_to_server(self):
        self.boot_and_verify_node()


class BaremetalRedfishIPxeWholediskHttpLink(
        bsm.BaremetalStandaloneScenarioTest):

    api_microversion = '1.31'  # to set the deploy_interface
    driver = 'redfish'
    deploy_interface = 'direct'
    boot_interface = 'ipxe'
    image_ref = CONF.baremetal.whole_disk_image_url
    image_checksum = CONF.baremetal.whole_disk_image_checksum
    wholedisk_image = True

    @decorators.idempotent_id('113acd0a-9872-4631-b3ee-54da7e3bb262')
    @utils.services('network')
    def test_ip_access_to_server(self):
        self.boot_and_verify_node()


class BaremetalRedfishDirectPxeWholedisk(
        bsm.BaremetalStandaloneScenarioTest):

    mandatory_attr = ['driver']
    api_microversion = '1.31'  # to set the deploy_interface
    driver = 'redfish'
    image_ref = CONF.baremetal.whole_disk_image_ref
    wholedisk_image = True
    boot_interface = 'pxe'
    deploy_interface = 'direct'

    @classmethod
    def skip_checks(cls):
        super(BaremetalRedfishDirectPxeWholedisk, cls).skip_checks()
        if CONF.baremetal_feature_enabled.ipxe_enabled:
            skip_msg = ("Skipping the test case since ipxe is enabled")
            raise cls.skipException(skip_msg)

    @utils.services('image', 'network')
    @decorators.idempotent_id('2bf34541-5514-4ea7-b073-7707d408ce99')
    def test_deploy_node(self):
        self.boot_and_verify_node()


class BaremetalIdracDirectIPxeWholedisk(
        bsm.BaremetalStandaloneScenarioTest):

    mandatory_attr = ['driver']
    api_microversion = '1.31'  # to set the deploy_interface
    driver = 'idrac'
    image_ref = CONF.baremetal.whole_disk_image_ref
    wholedisk_image = True
    boot_interface = 'ipxe'
    deploy_interface = 'direct'

    @classmethod
    def skip_checks(cls):
        super(BaremetalIdracDirectIPxeWholedisk, cls).skip_checks()
        if not CONF.baremetal_feature_enabled.ipxe_enabled:
            skip_msg = ("Skipping the test case since ipxe is disabled")
            raise cls.skipException(skip_msg)

    @utils.services('image', 'network')
    @decorators.idempotent_id('b467889b-aba4-4696-9e6f-4ef60a3f78a4')
    def test_deploy_node(self):
        self.boot_and_verify_node()


class BaremetalIdracDirectPxeWholedisk(
        bsm.BaremetalStandaloneScenarioTest):

    mandatory_attr = ['driver']
    api_microversion = '1.31'  # to set the deploy_interface
    driver = 'idrac'
    image_ref = CONF.baremetal.whole_disk_image_ref
    wholedisk_image = True
    boot_interface = 'pxe'
    deploy_interface = 'direct'

    @classmethod
    def skip_checks(cls):
        super(BaremetalIdracDirectPxeWholedisk, cls).skip_checks()
        if CONF.baremetal_feature_enabled.ipxe_enabled:
            skip_msg = ("Skipping the test case since ipxe is enabled")
            raise cls.skipException(skip_msg)

    @utils.services('image', 'network')
    @decorators.idempotent_id('20c0f2f4-723b-4d1e-a4f2-bea11ffb1513')
    def test_deploy_node(self):
        self.boot_and_verify_node()


class BaremetalIdracWSManDirectIPxeWholedisk(
        BaremetalIdracDirectIPxeWholedisk):
    power_interface = 'idrac-wsman'
    management_interface = 'idrac-wsman'


class BaremetalIdracWSManDirectPxeWholedisk(
        BaremetalIdracDirectPxeWholedisk):
    power_interface = 'idrac-wsman'
    management_interface = 'idrac-wsman'


class BaremetalIdracRedfishDirectIPxeWholedisk(
        BaremetalIdracDirectIPxeWholedisk):

    power_interface = 'idrac-redfish'
    management_interface = 'idrac-redfish'


class BaremetalIdracRedfishDirectPxeWholedisk(
        BaremetalIdracDirectPxeWholedisk):

    power_interface = 'idrac-redfish'
    management_interface = 'idrac-redfish'


class BaremetalIpmiDirectIPxeWholedisk(
        bsm.BaremetalStandaloneScenarioTest):

    mandatory_attr = ['driver']
    api_microversion = '1.31'  # to set the deploy_interface
    driver = 'ipmi'
    image_ref = CONF.baremetal.whole_disk_image_ref
    wholedisk_image = True
    boot_interface = 'ipxe'
    deploy_interface = 'direct'

    @classmethod
    def skip_checks(cls):
        super(BaremetalIpmiDirectIPxeWholedisk, cls).skip_checks()
        if not CONF.baremetal_feature_enabled.ipxe_enabled:
            skip_msg = ("Skipping the test case since ipxe is disabled")
            raise cls.skipException(skip_msg)

    @utils.services('image', 'network')
    @decorators.idempotent_id('9cd5a47b-ef96-4d64-9118-be467b24d1bf')
    def test_deploy_node(self):
        self.boot_and_verify_node()


class BaremetalIpmiDirectPxeWholedisk(
        bsm.BaremetalStandaloneScenarioTest):

    mandatory_attr = ['driver']
    api_microversion = '1.31'  # to set the deploy_interface
    driver = 'ipmi'
    image_ref = CONF.baremetal.whole_disk_image_ref
    wholedisk_image = True
    boot_interface = 'pxe'
    deploy_interface = 'direct'

    @classmethod
    def skip_checks(cls):
        super(BaremetalIpmiDirectPxeWholedisk, cls).skip_checks()
        if CONF.baremetal_feature_enabled.ipxe_enabled:
            skip_msg = ("Skipping the test case since ipxe is enabled")
            raise cls.skipException(skip_msg)

    @utils.services('image', 'network')
    @decorators.idempotent_id('d68a38aa-b731-403a-9d40-b3b49ea75e9b')
    def test_deploy_node(self):
        self.boot_and_verify_node()


class BaremetalIdracVirtualMediaWholedisk(
        bsm.BaremetalStandaloneScenarioTest):

    mandatory_attr = ['driver', 'boot_interface']
    api_microversion = '1.31'   # to set the deploy_interface
    driver = 'idrac'
    boot_interface = 'idrac-redfish-virtual-media'
    image_ref = CONF.baremetal.whole_disk_image_ref
    wholedisk_image = True
    deploy_interface = 'direct'

    @decorators.idempotent_id('b0bc87a5-4324-4134-bd5f-4bb1cf549e5c')
    @utils.services('image', 'network')
    def test_deploy_virtual_media_boot(self):
        self.boot_and_verify_node()


class BaremetalIdracSyncBootModeDirectWholedisk(
        bsm.BaremetalStandaloneScenarioTest):

    mandatory_attr = ['driver', 'power_interface', 'management_interface',
                      'bios_interface']

    api_microversion = '1.40'  # to get list_node_bios_settings
    driver = 'idrac'

    # NOTE: Image with UEFI support is necessary
    image_ref = CONF.baremetal.whole_disk_image_ref
    wholedisk_image = True
    boot_interface = 'ipxe'
    power_interface = 'idrac-redfish'
    management_interface = 'idrac-redfish'
    bios_interface = 'idrac-redfish'
    deploy_interface = 'direct'
    bios_mode = {'boot_mode': 'bios'}
    uefi_mode = {'boot_mode': 'uefi'}

    def _get_boot_mode_to_update(self):
        _, bios_settings = self.baremetal_client.\
            list_node_bios_settings(self.node['uuid'])
        for attr_name in bios_settings['bios']:
            if attr_name['name'] == 'BootMode':
                if attr_name['value'] == 'Uefi':
                    boot_mode = self.bios_mode
                else:
                    boot_mode = self.uefi_mode
        return boot_mode

    @decorators.idempotent_id('c2bebda2-fd27-4b10-9015-f7d877f0eb60')
    @utils.services('image', 'network')
    def test_sync_boot_mode(self):
        boot_mode_path = '/instance_info/capabilities'
        mode_to_update = self._get_boot_mode_to_update()
        self.update_node(self.node['uuid'], [{'op': 'replace',
                                              'path': boot_mode_path,
                                              'value': mode_to_update}])
        self.boot_and_verify_node()
        if mode_to_update == self.uefi_mode:
            mode_to_update = self.bios_mode
        else:
            mode_to_update = self.uefi_mode
        self.update_node(self.node['uuid'], [{'op': 'replace',
                                              'path': boot_mode_path,
                                              'value': mode_to_update}])
        self.set_node_provision_state(self.node['uuid'], 'rebuild')
        self.wait_provisioning_state(self.node['uuid'], 'active',
                                     timeout=CONF.baremetal.active_timeout,
                                     interval=30)


class BaremetalRedfishIPxeAnacondaNoGlance(
        bsm.BaremetalStandaloneScenarioTest):

    api_microversion = '1.78'  # to set the deploy_interface
    driver = 'redfish'
    deploy_interface = 'anaconda'
    boot_interface = 'ipxe'
    image_ref = CONF.baremetal.anaconda_image_ref
    wholedisk_image = None

    @classmethod
    def skip_checks(cls):
        super(BaremetalRedfishIPxeAnacondaNoGlance, cls).skip_checks()
        if 'anaconda' not in CONF.baremetal.enabled_deploy_interfaces:
            skip_msg = ("Skipping the test case since anaconda is not "
                        "enabled.")
            raise cls.skipException(skip_msg)

    def test_ip_access_to_server_without_stage2(self):
        # Tests deploy from a URL, or a pre-existing anaconda reference in
        # glance.
        if CONF.baremetal.anaconda_stage2_ramdisk_ref is not None:
            skip_msg = ("Skipping the test case as an anaconda stage2 "
                        "ramdisk has been defined, and that test can "
                        "run instead.")
            raise self.skipException(skip_msg)

        self.boot_and_verify_anaconda_node(
            image_ref=self.image_ref,
            kernel_ref=CONF.baremetal.anaconda_kernel_ref,
            ramdisk_ref=CONF.baremetal.anaconda_initial_ramdisk_ref)

    def test_ip_access_to_server_using_stage2(self):
        # Tests anaconda with a second stage ramdisk
        if CONF.baremetal.anaconda_stage2_ramdisk_ref is None:
            skip_msg = ("Skipping as stage2 ramdisk ref pointer has "
                        "not been configured.")
            raise self.skipException(skip_msg)

        self.boot_and_verify_anaconda_node(
            image_ref=self.image_ref,
            kernel_ref=CONF.baremetal.anaconda_kernel_ref,
            ramdisk_ref=CONF.baremetal.anaconda_initial_ramdisk_ref,
            stage2_ref=CONF.baremetal.anaconda_stage2_ramdisk_ref)


class BaremetalChangeBaseBootInterface(
        bsm.BaremetalStandaloneScenarioTest):

    # NOTE(TheJulia): The resource setup class *already* auto-sets the
    # requested boot interface. We just never tried to articulate
    # this sort of configuration before this class.
    # The goal: Test as many boot interfaces to ensure they are functional
    # under the base context

    api_microversion = '1.31'  # to set the boot_interface
    driver = None
    # This is intentionally meant to be a relatively light weight test,
    # but one that covers the common cases, i.e. direct is most appropriate
    # here. Ramdisk/anaconda/ansible and friends are all more specific
    # cased.
    deploy_interface = 'direct'
    image_ref = CONF.baremetal.whole_disk_image_url
    image_checksum = CONF.baremetal.whole_disk_image_checksum
    wholedisk_image = True

    # This is a special override which allows us to just use the base
    # driver interface defined on the node, and not have to special case it
    # and focus on the "driver".
    use_available_driver = True

    # List of valid drivers which these tests *can* attempt to utilize.
    # Generally these should be the most common, stock, upstream drivers.
    valid_driver_list = ['ipmi', 'redfish']

    # Bypass secondary attribute presence check as these tests don't require
    # the driver to be set.
    mandatory_attr = ['image_ref']


class BaremetalPXEBootTestClass(BaremetalChangeBaseBootInterface):

    boot_interface = 'pxe'

    @decorators.idempotent_id('62c12d2c-8c9f-4526-b9ab-b9cd63e0ea8a')
    @utils.services('network')
    def test_ip_access_to_server(self):
        self.boot_and_verify_node()


class BaremetalIPXEBootTestClass(BaremetalChangeBaseBootInterface):

    boot_interface = 'ipxe'

    @decorators.idempotent_id('113acd0a-9872-4631-b3ee-54da7e3bb262')
    @utils.services('network')
    def test_ip_access_to_server(self):
        self.boot_and_verify_node()


class BaremetalHTTPBootTestClass(BaremetalChangeBaseBootInterface):

    boot_interface = 'http'

    @decorators.idempotent_id('782c43db-77a1-4a3a-b46e-0ce9cbb7fba5')
    @utils.services('network')
    def test_ip_access_to_server(self):
        self.boot_and_verify_node()


class BaremetalHttpIPXEBootTestClass(BaremetalChangeBaseBootInterface):

    boot_interface = 'http-ipxe'

    @decorators.idempotent_id('45400d8e-55a5-4ba6-81f5-935a4183ed90')
    @utils.services('network')
    def test_ip_access_to_server(self):
        self.boot_and_verify_node()


class BaremetalRedfishVmediaBootTestClass(BaremetalChangeBaseBootInterface):

    boot_interface = 'redfish-virtual-media'
    driver = 'redfish'

    @decorators.idempotent_id('10535270-27e5-4616-9013-9507b1960dfa')
    @utils.services('network')
    def test_ip_access_to_server(self):
        self.boot_and_verify_node()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import functools
from http import client as http_client
from urllib import parse as urllib_parse

from oslo_serialization import jsonutils as json
from tempest.lib.common import api_version_utils
from tempest.lib.common import rest_client

# NOTE(vsaienko): concurrent tests work because they are launched in
# separate processes so global variables are not shared among them.
BAREMETAL_MICROVERSION = None

# Interfaces that can be set via the baremetal client and by logic in scenario
# managers.
SUPPORTED_INTERFACES = ['bios', 'deploy', 'rescue', 'boot', 'raid',
                        'management', 'power', 'inspect']


def set_baremetal_api_microversion(baremetal_microversion):
    global BAREMETAL_MICROVERSION
    BAREMETAL_MICROVERSION = baremetal_microversion


def reset_baremetal_api_microversion():
    global BAREMETAL_MICROVERSION
    BAREMETAL_MICROVERSION = None


def handle_errors(f):
    """A decorator that allows to ignore certain types of errors."""

    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        param_name = 'ignore_errors'
        ignored_errors = kwargs.get(param_name, tuple())

        if param_name in kwargs:
            del kwargs[param_name]

        try:
            return f(*args, **kwargs)
        except ignored_errors:
            # Silently ignore errors
            pass

    return wrapper


class BaremetalClient(rest_client.RestClient):
    """Base Tempest REST client for Ironic API."""

    api_microversion_header_name = 'X-OpenStack-Ironic-API-Version'
    uri_prefix = ''

    def get_headers(self):
        headers = super(BaremetalClient, self).get_headers()
        if BAREMETAL_MICROVERSION:
            # NOTE(TheJulia): This is not great, because it can blind a test
            # to the actual version supported.
            headers[self.api_microversion_header_name] = BAREMETAL_MICROVERSION
        return headers

    def get_raw_headers(self):
        """A proper get headers without guessing the microversion."""
        return super(BaremetalClient, self).get_headers()

    def get_min_max_api_microversions(self):
        """Returns a tuple of minimum and remote microversions."""
        if '/v1' in self.base_url:
            root_uri = '/'
        else:
            # NOTE(dtantsur): we should just use / here but due to a bug in
            # Ironic, / does not contain the microversion headers. See
            # https://bugs.launchpad.net/ironic/+bug/2079023
            root_uri = '/v1'
        _, resp_body = self._show_request(None, uri=root_uri)
        try:
            version = resp_body['default_version']
        except KeyError:
            version = resp_body['version']
        api_min = version.get('min_version')
        api_max = version.get('version')
        return (api_min, api_max)

    def request(self, *args, **kwargs):
        resp, resp_body = super(BaremetalClient, self).request(*args, **kwargs)
        latest_microversion = api_version_utils.LATEST_MICROVERSION
        if (BAREMETAL_MICROVERSION
                and BAREMETAL_MICROVERSION != latest_microversion):
            api_version_utils.assert_version_header_matches_request(
                self.api_microversion_header_name,
                BAREMETAL_MICROVERSION,
                resp)
        return resp, resp_body

    def serialize(self, object_dict):
        """Serialize an Ironic object."""

        return json.dumps(object_dict)

    def deserialize(self, object_str):
        """Deserialize an Ironic object."""

        return json.loads(object_str)

    def _get_uri(self, resource_name, uuid=None, permanent=False,
                 params=None):
        """Get URI for a specific resource or object.

        :param resource_name: The name of the REST resource, e.g., 'nodes'.
        :param uuid: The unique identifier of an object in UUID format.
        :returns: Relative URI for the resource or object.

        """
        prefix = self.uri_prefix if not permanent else ''
        if params:
            params = '?' + '&'.join('%s=%s' % tpl for tpl in params.items())
        else:
            params = ''

        return '{pref}/{res}{uuid}{params}'.format(
            pref=prefix, res=resource_name,
            uuid='/%s' % uuid if uuid else '',
            params=params)

    def _make_patch(self, allowed_attributes, **kwargs):
        """Create a JSON patch according to RFC 6902.

        :param allowed_attributes: An iterable object that contains a set of
            allowed attributes for an object.
        :param **kwargs: Attributes and new values for them.
        :returns: A JSON path that sets values of the specified attributes to
            the new ones.

        """
        def get_change(kwargs, path='/'):
            for name, value in kwargs.items():
                if isinstance(value, dict):
                    for ch in get_change(value, path + '%s/' % name):
                        yield ch
                else:
                    if value is None:
                        yield {'path': path + name,
                               'op': 'remove'}
                    else:
                        yield {'path': path + name,
                               'value': value,
                               'op': 'replace'}

        patch = [ch for ch in get_change(kwargs)
                 if ch['path'].lstrip('/') in allowed_attributes]

        return patch

    def _list_request(self, resource, permanent=False, headers=None,
                      extra_headers=False, **kwargs):
        """Get the list of objects of the specified type.

        :param resource: The name of the REST resource, e.g., 'nodes'.
        :param headers: List of headers to use in request.
        :param extra_headers: Specify whether to use headers.
        :param **kwargs: Parameters for the request.
        :returns: A tuple with the server response and deserialized JSON list
                 of objects

        """
        uri = self._get_uri(resource, permanent=permanent)
        if kwargs:
            uri += "?%s" % urllib_parse.urlencode(kwargs)

        resp, body = self.get(uri, headers=headers,
                              extra_headers=extra_headers)
        self.expected_success(http_client.OK, resp.status)

        return resp, self.deserialize(body)

    def _show_request(self,
                      resource,
                      uuid=None,
                      permanent=False,
                      headers=None,
                      extra_headers=False,
                      **kwargs):
        """Gets a specific object of the specified type.

        :param uuid: Unique identifier of the object in UUID format.
        :param headers: List of headers to use in request.
        :param extra_headers: Specify whether to use headers.
        :returns: Serialized object as a dictionary.

        """
        if 'uri' in kwargs:
            uri = kwargs['uri']
        else:
            uri = self._get_uri(resource, uuid=uuid, permanent=permanent)
        resp, body = self.get(uri, headers=headers,
                              extra_headers=extra_headers)
        self.expected_success(http_client.OK, resp.status)

        return resp, self.deserialize(body)

    def _create_request(self, resource, object_dict):
        """Create an object of the specified type.

        :param resource: The name of the REST resource, e.g., 'nodes'.
        :param object_dict: A Python dict that represents an object of the
                            specified type.
        :returns: A tuple with the server response and the deserialized created
                 object.

        """
        body = self.serialize(object_dict)
        uri = self._get_uri(resource)

        resp, body = self.post(uri, body=body)
        self.expected_success(http_client.CREATED, resp.status)

        return resp, self.deserialize(body)

    def _create_request_no_response_body(self, resource, object_dict):
        """Create an object of the specified type.

           Do not expect any body in the response.

        :param resource: The name of the REST resource, e.g., 'nodes'.
        :param object_dict: A Python dict that represents an object of the
                            specified type.
        :returns: The server response.
        """

        body = self.serialize(object_dict)
        uri = self._get_uri(resource)

        resp, body = self.post(uri, body=body)
        self.expected_success(http_client.NO_CONTENT, resp.status)

        return resp

    def _delete_request(self, resource, uuid,
                        expected_status=http_client.NO_CONTENT):
        """Delete specified object.

        :param resource: The name of the REST resource, e.g., 'nodes'.
        :param uuid: The unique identifier of an object in UUID format.
        :param expected_status: Expected response status code. By default is
                                http_client.NO_CONTENT (204)
        :returns: A tuple with the server response and the response body.

        """
        uri = self._get_uri(resource, uuid)

        resp, body = self.delete(uri)
        self.expected_success(expected_status, resp.status)
        return resp, body

    def _patch_request(self, resource, uuid, patch_object, params=None):
        """Update specified object with JSON-patch.

        :param resource: The name of the REST resource, e.g., 'nodes'.
        :param uuid: The unique identifier of an object in UUID format.
        :param params: query parameters to pass.
        :returns: A tuple with the server response and the serialized patched
                 object.

        """
        uri = self._get_uri(resource, uuid, params=params)
        patch_body = json.dumps(patch_object)

        resp, body = self.patch(uri, body=patch_body)
        self.expected_success(http_client.OK, resp.status)
        return resp, self.deserialize(body)

    @handle_errors
    def get_api_description(self):
        """Retrieves all versions of the Ironic API."""

        return self._list_request('', permanent=True)

    @handle_errors
    def get_version_description(self, version='v1'):
        """Retrieves the description of the API.

        :param version: The version of the API. Default: 'v1'.
        :returns: Serialized description of API resources.

        """
        return self._list_request(version, permanent=True)

    def _put_request(self, resource, put_object):
        """Update specified object with JSON-patch."""
        uri = self._get_uri(resource)
        put_body = json.dumps(put_object)

        resp, body = self.put(uri, body=put_body)
        self.expected_success([http_client.ACCEPTED, http_client.NO_CONTENT],
                              resp.status)
        return resp, body



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

from tempest.lib.common.utils import data_utils
from tempest.lib import decorators
from tempest.lib import exceptions as lib_exc

from ironic_tempest_plugin.tests.api.admin import api_microversion_fixture
from ironic_tempest_plugin.tests.api import base


EXAMPLE_STEPS = [{
    'interface': 'bios',
    'step': 'apply_configuration',
    'args': {},
    'order': 1
}]


def _get_random_trait():
    return data_utils.rand_name('CUSTOM', '').replace('-', '_')


class TestRunbooks(base.BaseBaremetalTest):
    """Tests for runbooks."""

    min_microversion = '1.92'

    def setUp(self):
        super(TestRunbooks, self).setUp()
        self.name = _get_random_trait()
        self.steps = copy.deepcopy(EXAMPLE_STEPS)
        _, self.runbook = self.create_runbook(self.name,
                                              steps=self.steps)

        _, self.chassis = self.create_chassis()
        _, self.test_node = self.create_node(self.chassis['uuid'])

    @decorators.idempotent_id('5a0cfe27-b7e8-d4f0-d3c5-e6e1eef63241')
    def test_create_runbook_specifying_uuid(self):
        name = _get_random_trait()
        uuid = data_utils.rand_uuid()

        _, runbook = self.create_runbook(name=name, steps=self.steps,
                                         uuid=uuid)

        _, body = self.client.show_runbook(uuid)
        self._assertExpected(runbook, body)

    @decorators.idempotent_id('b8bfb388-97b0-aa6e-5130-1a1ed34f94db')
    def test_delete_runbook(self):
        self.delete_runbook(self.runbook['uuid'])
        self.assertRaises(lib_exc.NotFound, self.client.show_runbook,
                          self.runbook['uuid'])

    @decorators.idempotent_id('f1cba93a-2894-7296-0f4b-58867359480b')
    def test_show_runbook(self):
        _, runbook = self.client.show_runbook(self.runbook['uuid'])
        self._assertExpected(self.runbook, runbook)
        self.assertEqual(self.name, runbook['name'])
        self.assertEqual(self.steps, runbook['steps'])
        self.assertIn('uuid', runbook)
        self.assertEqual({}, runbook['extra'])

    @decorators.idempotent_id('c95e2631-24e2-914b-db00-c8dafc35a677')
    def test_show_runbook_with_links(self):
        _, runbook = self.client.show_runbook(self.runbook['uuid'])
        self.assertIn('links', runbook)
        self.assertEqual(2, len(runbook['links']))
        self.assertIn(runbook['uuid'], runbook['links'][0]['href'])

    @decorators.idempotent_id('7b7951b3-e177-21d7-933a-1f29891dea52')
    def test_list_runbooks(self):
        _, body = self.client.list_runbooks()
        self.assertIn(self.runbook['uuid'],
                      [i['uuid'] for i in body['runbooks']])

        for runbook in body['runbooks']:
            self.validate_self_link('runbooks', runbook['uuid'],
                                    runbook['links'][0]['href'])

    @decorators.idempotent_id('6aafc619-0d98-5341-7d94-b293e194dcf7')
    def test_list_with_limit(self):
        for _ in range(2):
            name = _get_random_trait()
            self.create_runbook(name, steps=self.steps)

        _, body = self.client.list_runbooks(limit=3)

        next_marker = body['runbooks'][-1]['uuid']
        self.assertIn(next_marker, body['next'])

    @decorators.idempotent_id('ebd762c3-c10e-b71f-5efd-af14ac9f6092')
    def test_list_runbooks_detail(self):
        uuids = [
            self.create_runbook(_get_random_trait(), steps=self.steps)
            [1]['uuid'] for _ in range(0, 5)]

        _, body = self.client.list_runbooks(detail=True)

        runbooks_dict = dict((runbook['uuid'], runbook)
                             for runbook in body['runbooks']
                             if runbook['uuid'] in uuids)

        for uuid in uuids:
            self.assertIn(uuid, runbooks_dict)
            runbook = runbooks_dict[uuid]
            self.assertIn('name', runbook)
            self.assertEqual(self.steps, runbook['steps'])
            self.assertIn('uuid', runbook)
            self.assertEqual({}, runbook['extra'])
            self.validate_self_link('runbooks', runbook['uuid'],
                                    runbook['links'][0]['href'])

    @decorators.idempotent_id('fb192fdd-ea6a-c637-a36e-390c46a7663b')
    def test_update_runbook_replace(self):
        new_name = _get_random_trait()
        new_steps = [{
            'interface': 'raid',
            'step': 'create_configuration',
            'args': {},
            'order': 2,
        }]

        patch = [{'path': '/name', 'op': 'replace', 'value': new_name},
                 {'path': '/steps', 'op': 'replace', 'value': new_steps}]

        self.client.update_runbook(self.runbook['uuid'], patch)

        _, body = self.client.show_runbook(self.runbook['uuid'])
        self.assertEqual(new_name, body['name'])
        self.assertEqual(new_steps, body['steps'])

    @decorators.idempotent_id('99f52546-9906-6ded-9186-7262591b99ec')
    def test_update_runbook_add(self):
        new_steps = [
            {
                'interface': 'bios',
                'step': 'cache_bios_settings',
                'args': {},
                'order': 2
            },
            {
                'interface': 'bios',
                'step': 'factory_reset',
                'args': {},
                'order': 3
            },
        ]

        patch = [{'path': '/steps/1', 'op': 'add', 'value': new_steps[0]},
                 {'path': '/steps/2', 'op': 'add', 'value': new_steps[1]}]

        self.client.update_runbook(self.runbook['uuid'], patch)

        _, body = self.client.show_runbook(self.runbook['uuid'])
        self.assertEqual(self.steps + new_steps, body['steps'])

    @decorators.idempotent_id('ec1550c3-264e-fcce-b131-d2815fdb733b')
    def test_update_runbook_mixed_ops(self):
        new_name = _get_random_trait()
        new_steps = [
            {
                'interface': 'bios',
                'step': 'apply_configuration',
                'args': {},
                'order': 2
            },
            {
                'interface': 'bios',
                'step': 'apply_configuration',
                'args': {},
                'order': 3
            },
        ]

        patch = [{'path': '/name', 'op': 'replace', 'value': new_name},
                 {'path': '/steps/0', 'op': 'replace', 'value': new_steps[0]},
                 {'path': '/steps/0', 'op': 'remove'},
                 {'path': '/steps/0', 'op': 'add', 'value': new_steps[1]}]

        self.client.update_runbook(self.runbook['uuid'], patch)

        _, body = self.client.show_runbook(self.runbook['uuid'])
        self.assertEqual(new_name, body['name'])
        self.assertEqual([new_steps[1]], body['steps'])

    @decorators.idempotent_id('5c7f0aca-cee3-d083-ef2a-e33a8dc467c5')
    def test_combining_runbook_and_explicit_steps(self):
        explicit_steps = [{'interface': 'deploy', 'step': 'deploy',
                           'args': {}}]

        self.assertRaises(lib_exc.BadRequest,
                          self.client.set_node_provision_state,
                          self.test_node['uuid'], 'active',
                          runbook=self.runbook['uuid'],
                          clean_steps=explicit_steps)

    @decorators.idempotent_id('ddf4a9b7-144d-386a-fd3fcf8960d77199')
    def test_create_runbook_with_invalid_step_format(self):
        name = _get_random_trait()
        invalid_steps = [{'invalid_key': 'value'}]

        self.assertRaises(lib_exc.BadRequest, self.create_runbook,
                          name, steps=invalid_steps)


class TestRunbooksOldAPI(base.BaseBaremetalTest):
    """Negative tests for runbooks using an old API version."""

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('e9481a0d-23e0-4757-bc11-c3c9ab9d3839')
    def test_create_runbook_old_api(self):
        # With runbooks support, ironic returns 404. Without, 405.
        self.assertRaises((lib_exc.NotFound, lib_exc.UnexpectedResponseCode),
                          self.create_runbook,
                          name=_get_random_trait(), steps=EXAMPLE_STEPS)

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('0d3af2aa-ba53-4c8a-92d4-91f9b4179fe7')
    def test_update_runbook_old_api(self):
        patch = [{'path': '/name', 'op': 'replace',
                  'value': _get_random_trait()}]

        # With runbooks support, ironic returns 404. Without, 405.
        self.assertRaises((lib_exc.NotFound, lib_exc.UnexpectedResponseCode),
                          self.client.update_runbook,
                          _get_random_trait(), patch)

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('1646b1e5-ab81-45a8-9ea0-30444a4dcaa2')
    def test_delete_runbook_old_api(self):
        # With runbooks support, ironic returns 404. Without, 405.
        self.assertRaises((lib_exc.NotFound, lib_exc.UnexpectedResponseCode),
                          self.client.delete_runbook,
                          _get_random_trait())

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('819480ac-f36a-4402-b1d5-504d7cf55b1f')
    def test_list_runbooks_old_api(self):
        self.assertRaises(lib_exc.NotFound,
                          self.client.list_runbooks)


class TestRunbooksNegative(base.BaseBaremetalTest):
    """Negative tests for runbooks."""

    min_microversion = '1.92'

    def setUp(self):
        super(TestRunbooksNegative, self).setUp()
        self.useFixture(
            api_microversion_fixture.APIMicroversionFixture(
                self.min_microversion)
        )
        self.steps = EXAMPLE_STEPS

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('a4085c08-e718-4c2f-a796-0e115b659243')
    def test_create_runbook_invalid_name(self):
        name = 'invalid-name'
        self.assertRaises(lib_exc.BadRequest,
                          self.create_runbook, name=name,
                          steps=self.steps)

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('6390acc4-9490-4b23-8b4c-41888a78c9b7')
    def test_create_runbook_duplicated_name(self):
        name = _get_random_trait()
        self.create_runbook(name=name, steps=self.steps)
        self.assertRaises(lib_exc.Conflict, self.create_runbook,
                          name=name, steps=self.steps)

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('ed3f0cec-13e8-4175-9fdb-d129e7b7fe10')
    def test_create_runbook_no_mandatory_field_name(self):
        self.assertRaises(lib_exc.BadRequest, self.create_runbook,
                          name=None, steps=self.steps)

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('af5dd0df-d903-463f-9535-9e4e9d6fd576')
    def test_create_runbook_no_mandatory_field_steps(self):
        self.assertRaises(lib_exc.BadRequest, self.create_runbook,
                          name=_get_random_trait())

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('cbd33bc5-7602-40b7-943e-3e92217567a3')
    def test_create_runbook_malformed_steps(self):
        steps = {'key': 'value'}
        self.assertRaises(lib_exc.BadRequest, self.create_runbook,
                          name=_get_random_trait(), steps=steps)

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('2a562fca-f377-4a6e-b332-37ee82d3a983')
    def test_create_runbook_malformed_runbook_uuid(self):
        uuid = 'malformed:uuid'
        self.assertRaises(lib_exc.BadRequest, self.create_runbook,
                          name=_get_random_trait(), steps=self.steps,
                          uuid=uuid)

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('2c006994-88ca-43b7-b605-897d479229d9')
    def test_show_runbook_nonexistent(self):
        self.assertRaises(lib_exc.NotFound, self.client.show_runbook,
                          data_utils.rand_uuid())

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('5a815f37-f015-4d68-9b22-099504f74805')
    def test_update_runbook_remove_mandatory_field_steps(self):
        name = _get_random_trait()
        _, runbook = self.create_runbook(name=name, steps=self.steps)

        self.assertRaises(lib_exc.BadRequest,
                          self.client.update_runbook,
                          runbook['uuid'],
                          [{'path': '/steps/0', 'op': 'remove'}])

        self.assertRaises(lib_exc.BadRequest,
                          self.client.update_runbook,
                          runbook['uuid'],
                          [{'path': '/steps', 'op': 'remove'}])

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('ee852ebb-a601-4593-9d59-063fcbc8f964')
    def test_update_runbook_remove_mandatory_field_name(self):
        name = _get_random_trait()
        _, runbook = self.create_runbook(name=name, steps=self.steps)
        self.assertRaises(lib_exc.BadRequest,
                          self.client.update_runbook,
                          runbook['uuid'],
                          [{'path': '/name', 'op': 'remove'}])

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('e59bf38d-272f-4490-b21e-9db217f11378')
    def test_update_runbook_replace_empty_name(self):
        name = _get_random_trait()
        _, runbook = self.create_runbook(name=name, steps=self.steps)
        self.assertRaises(lib_exc.BadRequest,
                          self.client.update_runbook,
                          runbook['uuid'],
                          [{'path': '/name', 'op': 'replace', 'value': ''}])



# Copyright 2014 NEC Corporation. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest import config
from tempest.lib import decorators

from ironic_tempest_plugin.tests.api.admin import api_microversion_fixture
from ironic_tempest_plugin.tests.api import base

CONF = config.CONF


class TestDrivers(base.BaseBaremetalTest):
    """Tests for drivers."""

    @classmethod
    def resource_setup(cls):
        super(TestDrivers, cls).resource_setup()
        cls.driver_name = CONF.baremetal.driver

    @decorators.idempotent_id('5aed2790-7592-4655-9b16-99abcc2e6ec5')
    def test_list_drivers(self):
        _, drivers = self.client.list_drivers()
        self.assertIn(self.driver_name,
                      [d['name'] for d in drivers['drivers']])

    @decorators.idempotent_id('fb3287a3-c4d7-44bf-ae9d-1eef906d78ce')
    def test_show_driver(self):
        _, driver = self.client.show_driver(self.driver_name)
        self.assertEqual(self.driver_name, driver['name'])

    @decorators.idempotent_id('6efa976f-78a2-4859-b3aa-97d960d6e5e5')
    def test_driver_properties(self):
        _, properties = self.client.get_driver_properties(self.driver_name)
        self.assertNotEmpty(properties)

    @decorators.idempotent_id('fdf61f5a-f59d-4235-ad6c-cc718740e3e3')
    def test_driver_logical_disk_properties(self):
        self.useFixture(
            api_microversion_fixture.APIMicroversionFixture('1.12'))
        _, properties = self.client.get_driver_logical_disk_properties(
            self.driver_name)
        self.assertNotEmpty(properties)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest import config
from tempest.lib import decorators
from tempest import test  # noqa

from ironic_tempest_plugin.tests.scenario import baremetal_manager
from ironic_tempest_plugin.tests.scenario import introspection_manager

CONF = config.CONF

ProvisionStates = baremetal_manager.BaremetalProvisionStates


class InspectorDiscoveryTest(introspection_manager.InspectorScenarioTest):
    @classmethod
    def skip_checks(cls):
        super(InspectorDiscoveryTest, cls).skip_checks()
        if not CONF.baremetal_introspection.auto_discovery_feature:
            msg = ("Please, provide a value for node_not_found_hook in "
                   "processing section of inspector.conf for enable "
                   "auto-discovery feature.")
            raise cls.skipException(msg)

    def setUp(self):
        super(InspectorDiscoveryTest, self).setUp()

        discovered_node = self._get_discovery_node()
        self.node_info = self._get_node_info(discovered_node)
        self.default_driver = (
            CONF.baremetal_introspection.auto_discovery_default_driver
        )
        self.expected_driver = (
            CONF.baremetal_introspection.auto_discovery_target_driver
            or CONF.baremetal_introspection.auto_discovery_default_driver
        )

        rule = {
            "description": "Auto-discovery rule",
            "actions": [
                # Set an attribute in case we cannot access the introspection
                # data to check the auto_discovered flag directly.
                {"action": "set-attribute",
                 "path": "/extra/discovered",
                 "value": "yes"},
                # Re-assign the name to use it later in the test.
                {"action": "set-attribute",
                 "path": "/name",
                 "value": self.node_info['name']},
            ],
            "conditions": [
                # This flag must be automatically set by the auto-discovery
                # process.
                {"op": "eq",
                 "field": "data://auto_discovered",
                 "value": True},
                # Making sure the initial driver matches the expected.
                {"op": "eq",
                 "field": "node://driver",
                 "value": self.default_driver},
            ]
        }

        if self.expected_driver != self.default_driver:
            rule['actions'].append({
                'action': 'set-attribute',
                'path': '/driver',
                'value': self.expected_driver
            })

        self.rule_import_from_dict(rule)
        self.addCleanup(self.rule_purge)

    def _get_node_info(self, node_uuid):
        node = self.node_show(node_uuid)
        ports = self.node_port_list(node_uuid)
        node['port_macs'] = [port['address'] for port in ports]
        return node

    def _get_discovery_node(self):
        nodes = self.node_list()

        discovered_node = None
        for node in nodes:
            if (node['provision_state'] == ProvisionStates.AVAILABLE
                    or node['provision_state'] == ProvisionStates.ENROLL
                    or node['provision_state'] is ProvisionStates.NOSTATE):
                discovered_node = node['uuid']
                break

        self.assertIsNotNone(discovered_node)
        return discovered_node

    def verify_node_introspection_data(self, node):
        data = self.introspection_data(node['uuid'])
        self.assertTrue(data['auto_discovered'])
        self.assertEqual(data['cpu_arch'],
                         self.flavor['properties']['cpu_arch'])
        self.assertEqual(int(data['memory_mb']),
                         int(self.flavor['ram']))
        self.assertEqual(int(data['cpus']), int(self.flavor['vcpus']))

    def verify_node_flavor(self, node):
        expected_cpus = self.flavor['vcpus']
        expected_memory_mb = self.flavor['ram']
        expected_cpu_arch = self.flavor['properties']['cpu_arch']
        disk_size = self.flavor['disk']
        ephemeral_size = self.flavor['OS-FLV-EXT-DATA:ephemeral']
        expected_local_gb = disk_size + ephemeral_size

        self.assertEqual(expected_cpus,
                         int(node['properties']['cpus']))
        self.assertEqual(expected_memory_mb,
                         int(node['properties']['memory_mb']))
        self.assertEqual(expected_local_gb,
                         int(node['properties']['local_gb']))
        self.assertEqual(expected_cpu_arch,
                         node['properties']['cpu_arch'])

    @decorators.idempotent_id('dd3abe5e-0d23-488d-bb4e-344cdeff7dcb')
    def test_bearmetal_auto_discovery(self):
        """This test case follows this set of operations:

           * Choose appropriate node, based on provision state;
           * Get node info;
           * Generate discovery rule;
           * Start introspection via ironic-inspector API;
           * Delete the node from ironic;
           * Waiting for node discovery;
           * Verify introspected node.
        """
        # NOTE(aarefiev): workaround for infra, 'tempest' user doesn't
        # have virsh privileges, so lets power on the node via ironic
        # and then delete it. Because of node is blacklisted in inspector
        # we can't just power on it, therefore start introspection is used
        # to whitelist discovered node first.
        self.baremetal_client.set_node_provision_state(
            self.node_info['uuid'], 'manage')
        self.introspection_start(self.node_info['uuid'])
        self.wait_power_state(
            self.node_info['uuid'],
            baremetal_manager.BaremetalPowerStates.POWER_ON)
        self.node_delete(self.node_info['uuid'])

        self.wait_for_node(self.node_info['name'])

        inspected_node = self.node_show(self.node_info['name'])
        self.verify_node_flavor(inspected_node)
        data_store = CONF.baremetal_introspection.data_store
        if data_store is None:
            # Backward compatibility, the option is not set.
            data_store = ('swift' if CONF.service_available.swift
                          else 'none')
        if data_store != 'none':
            self.verify_node_introspection_data(inspected_node)
        self.assertEqual(ProvisionStates.ENROLL,
                         inspected_node['provision_state'])
        self.assertEqual(self.expected_driver, inspected_node['driver'])
        self.assertEqual('yes', inspected_node['extra']['discovered'])



# Copyright 2015 NEC Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import os

from tempest import config
from tempest.test_discover import plugins

from ironic_tempest_plugin import config as project_config

_opts = [
    (project_config.baremetal_group, project_config.BaremetalGroup),
    (project_config.baremetal_features_group,
     project_config.BaremetalFeaturesGroup),
    (project_config.baremetal_introspection_group,
     project_config.BaremetalIntrospectionGroup),
]


class IronicTempestPlugin(plugins.TempestPlugin):
    def load_tests(self):
        base_path = os.path.split(os.path.dirname(
            os.path.abspath(__file__)))[0]
        test_dir = "ironic_tempest_plugin/tests"
        full_test_dir = os.path.join(base_path, test_dir)
        return full_test_dir, base_path

    def register_opts(self, conf):
        conf.register_opt(project_config.ironic_service_option,
                          group='service_available')
        conf.register_opt(project_config.inspector_service_option,
                          group='service_available')
        conf.register_opt(project_config.ironic_scope_enforcement,
                          group='enforce_scope')
        conf.register_opt(project_config.inspector_scope_enforcement,
                          group='enforce_scope')
        for group, option in _opts:
            config.register_opt_group(conf, group, option)

    def get_opt_lists(self):
        return [(group.name, option) for group, option in _opts]

    def get_service_clients(self):
        ironic_config = config.service_client_config(
            project_config.baremetal_group.name
        )
        baremetal_client = {
            'name': 'baremetal',
            'service_version': 'baremetal.v1',
            'module_path': 'ironic_tempest_plugin.services.baremetal.v1.'
                           'json.baremetal_client',
            'client_names': [
                'BaremetalClient',
            ],
        }
        baremetal_client.update(ironic_config)

        inspector_config = config.service_client_config(
            project_config.baremetal_introspection_group.name
        )
        inspector_client = {
            'name': 'introspection',
            'service_version': 'baremetal_introspection.v1',
            'module_path': 'ironic_tempest_plugin.services.'
                           'introspection_client',
            'client_names': [
                'BaremetalIntrospectionClient',
            ],
        }
        inspector_client.update(inspector_config)

        return [baremetal_client, inspector_client]



# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from tempest.lib.common.utils import data_utils
from tempest.lib import decorators
from tempest.lib import exceptions as lib_exc

from ironic_tempest_plugin.tests.api.admin import api_microversion_fixture
from ironic_tempest_plugin.tests.api import base


class TestVolumeConnector(base.BaseBaremetalTest):
    """Basic test cases for volume connector."""

    min_microversion = '1.32'
    extra = {'key1': 'value1', 'key2': 'value2'}

    def setUp(self):
        super(TestVolumeConnector, self).setUp()
        self.useFixture(
            api_microversion_fixture.APIMicroversionFixture(
                self.min_microversion))
        _, self.chassis = self.create_chassis()
        _, self.node = self.create_node(self.chassis['uuid'])
        _, self.volume_connector = self.create_volume_connector(
            self.node['uuid'], type='iqn',
            connector_id=data_utils.rand_name('connector_id'),
            extra=self.extra)

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('3c3cbf45-488a-4386-a811-bf0aa2589c58')
    def test_create_volume_connector_error(self):
        """Create a volume connector.

        Fail when creating a volume connector with same connector_id
        & type as an existing volume connector.
        """
        regex_str = (r'.*A volume connector .*already exists')

        self.assertRaisesRegex(
            lib_exc.Conflict, regex_str,
            self.create_volume_connector,
            self.node['uuid'],
            type=self.volume_connector['type'],
            connector_id=self.volume_connector['connector_id'])

    @decorators.idempotent_id('5795f816-0789-42e6-bb9c-91b4876ad13f')
    def test_delete_volume_connector(self):
        """Delete a volume connector."""
        # Powering off the Node before deleting a volume connector.
        self.client.set_node_power_state(self.node['uuid'], 'power off')

        self.delete_volume_connector(self.volume_connector['uuid'])
        self.assertRaises(lib_exc.NotFound, self.client.show_volume_connector,
                          self.volume_connector['uuid'])

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('ccbda5e6-52b7-400c-94d7-25eec1d590f0')
    def test_delete_volume_connector_error(self):
        """Delete a volume connector

        Fail when deleting a volume connector on node
        with powered on state.
        """

        # Powering on the Node before deleting a volume connector.
        self.client.set_node_power_state(self.node['uuid'], 'power on')

        regex_str = (r'.*The requested action \\\\"volume connector '
                     r'deletion\\\\" can not be performed on node*')

        self.assertRaisesRegex(lib_exc.BadRequest,
                               regex_str,
                               self.delete_volume_connector,
                               self.volume_connector['uuid'])

    @decorators.idempotent_id('6e4f50b7-0f4f-41c2-971e-d751abcac4e0')
    def test_show_volume_connector(self):
        """Show a specified volume connector."""
        _, volume_connector = self.client.show_volume_connector(
            self.volume_connector['uuid'])
        self._assertExpected(self.volume_connector, volume_connector)

    @decorators.idempotent_id('a4725778-e164-4ee5-96a0-66119a35f783')
    def test_list_volume_connectors(self):
        """List volume connectors."""
        _, body = self.client.list_volume_connectors()
        self.assertIn(self.volume_connector['uuid'],
                      [i['uuid'] for i in body['connectors']])
        self.assertIn(self.volume_connector['type'],
                      [i['type'] for i in body['connectors']])
        self.assertIn(self.volume_connector['connector_id'],
                      [i['connector_id'] for i in body['connectors']])

    @decorators.idempotent_id('1d0459ad-01c0-46db-b930-7301bc2a3c98')
    def test_list_with_limit(self):
        """List volume connectors with limit."""
        for i in range(2):
            _, self.volume_connector = self.create_volume_connector(
                self.node['uuid'], type='iqn',
                connector_id=data_utils.rand_name('connector_id'),
                extra=self.extra)

        _, body = self.client.list_volume_connectors(limit=3)

        next_marker = body['connectors'][-1]['uuid']
        self.assertIn(next_marker, body['next'])

    @decorators.idempotent_id('3c6f8354-e9bd-4f21-aae2-6deb96b04be7')
    def test_update_volume_connector_replace(self):
        """Update a volume connector with new connector id."""
        new_connector_id = data_utils.rand_name('connector_id')

        patch = [{'path': '/connector_id',
                  'op': 'replace',
                  'value': new_connector_id}]

        # Powering off the Node before updating a volume connector.
        self.client.set_node_power_state(self.node['uuid'], 'power off')

        self.client.update_volume_connector(
            self.volume_connector['uuid'], patch)

        _, body = self.client.show_volume_connector(
            self.volume_connector['uuid'])
        self.assertEqual(new_connector_id, body['connector_id'])

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('5af8dc7a-9965-4787-8184-e60aeaf30957')
    def test_update_volume_connector_replace_error(self):
        """Updating a volume connector.

        Fail when updating a volume connector on node
        with power on state.
        """

        new_connector_id = data_utils.rand_name('connector_id')

        patch = [{'path': '/connector_id',
                  'op': 'replace',
                  'value': new_connector_id}]

        # Powering on the Node before updating a volume connector.
        self.client.set_node_power_state(self.node['uuid'], 'power on')

        regex_str = (r'.*The requested action \\\\"volume connector '
                     r'update\\\\" can not be performed on node*')
        self.assertRaisesRegex(lib_exc.BadRequest,
                               regex_str,
                               self.client.update_volume_connector,
                               self.volume_connector['uuid'],
                               patch)

    @decorators.idempotent_id('b95c75eb-4048-482e-99ff-fe1d32538383')
    def test_update_volume_connector_remove_item(self):
        """Update a volume connector by removing one item from collection."""
        new_extra = {'key1': 'value1'}
        _, body = self.client.show_volume_connector(
            self.volume_connector['uuid'])
        connector_id = body['connector_id']
        connector_type = body['type']

        # Powering off the Node before updating a volume connector.
        self.client.set_node_power_state(self.node['uuid'], 'power off')

        # Removing one item from the collection
        self.client.update_volume_connector(self.volume_connector['uuid'],
                                            [{'path': '/extra/key2',
                                              'op': 'remove'}])
        _, body = self.client.show_volume_connector(
            self.volume_connector['uuid'])
        self.assertEqual(new_extra, body['extra'])

        # Assert nothing else was changed
        self.assertEqual(connector_id, body['connector_id'])
        self.assertEqual(connector_type, body['type'])

    @decorators.idempotent_id('8de03acd-532a-476f-8bc9-0e8b23bfe609')
    def test_update_volume_connector_remove_collection(self):
        """Update a volume connector by removing collection."""
        _, body = self.client.show_volume_connector(
            self.volume_connector['uuid'])
        connector_id = body['connector_id']
        connector_type = body['type']

        # Powering off the Node before updating a volume connector.
        self.client.set_node_power_state(self.node['uuid'], 'power off')

        # Removing the collection
        self.client.update_volume_connector(self.volume_connector['uuid'],
                                            [{'path': '/extra',
                                              'op': 'remove'}])
        _, body = self.client.show_volume_connector(
            self.volume_connector['uuid'])
        self.assertEqual({}, body['extra'])

        # Assert nothing else was changed
        self.assertEqual(connector_id, body['connector_id'])
        self.assertEqual(connector_type, body['type'])

    @decorators.idempotent_id('bfb0ca6b-086d-4663-9b25-e0eaf42da55b')
    def test_update_volume_connector_add(self):
        """Update a volume connector by adding one item to collection."""
        new_extra = {'key1': 'value1', 'key2': 'value2', 'key3': 'value3'}

        patch = [{'path': '/extra/key3',
                  'op': 'add',
                  'value': new_extra['key3']},
                 {'path': '/extra/key3',
                  'op': 'add',
                  'value': new_extra['key3']}]

        # Powering off the Node before updating a volume connector.
        self.client.set_node_power_state(self.node['uuid'], 'power off')

        self.client.update_volume_connector(
            self.volume_connector['uuid'], patch)

        _, body = self.client.show_volume_connector(
            self.volume_connector['uuid'])
        self.assertEqual(new_extra, body['extra'])



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import uuidutils
from tempest import config
from tempest.lib.common.utils import data_utils
from tempest.lib import decorators
from tempest.lib import exceptions as lib_exc

from ironic_tempest_plugin.common import waiters
from ironic_tempest_plugin.tests.api.admin import api_microversion_fixture
from ironic_tempest_plugin.tests.api import base

CONF = config.CONF


class TestNodes(base.BaseBaremetalTest):
    """Tests for baremetal nodes."""

    def setUp(self):
        super(TestNodes, self).setUp()

        _, self.chassis = self.create_chassis()
        _, self.node = self.create_node(self.chassis['uuid'])

    def _associate_node_with_instance(self):
        self.client.set_node_power_state(self.node['uuid'], 'power off')
        waiters.wait_for_bm_node_status(self.client, self.node['uuid'],
                                        'power_state', 'power off')
        instance_uuid = data_utils.rand_uuid()
        self.client.update_node(self.node['uuid'],
                                instance_uuid=instance_uuid)
        self.addCleanup(self.client.update_node,
                        uuid=self.node['uuid'], instance_uuid=None)
        return instance_uuid

    @decorators.idempotent_id('4e939eb2-8a69-4e84-8652-6fffcbc9db8f')
    def test_create_node(self):
        params = {'cpu_arch': 'x86_64',
                  'cpus': '12',
                  'local_gb': '10',
                  'memory_mb': '1024'}

        _, body = self.create_node(self.chassis['uuid'], **params)
        self._assertExpected(params, body['properties'])

    @decorators.idempotent_id('9ade60a4-505e-4259-9ec4-71352cbbaf47')
    def test_delete_node(self):
        _, node = self.create_node(self.chassis['uuid'])

        self.delete_node(node['uuid'])

        self.assertRaises(lib_exc.NotFound, self.client.show_node,
                          node['uuid'])

    @decorators.idempotent_id('55451300-057c-4ecf-8255-ba42a83d3a03')
    def test_show_node(self):
        _, loaded_node = self.client.show_node(self.node['uuid'])
        self._assertExpected(self.node, loaded_node)

    @decorators.idempotent_id('4ca123c4-160d-4d8d-a3f7-15feda812263')
    def test_list_nodes(self):
        _, body = self.client.list_nodes()
        self.assertIn(self.node['uuid'],
                      [i['uuid'] for i in body['nodes']])

    @decorators.idempotent_id('85b1f6e0-57fd-424c-aeff-c3422920556f')
    def test_list_nodes_association(self):
        _, body = self.client.list_nodes(associated=True)
        self.assertNotIn(self.node['uuid'],
                         [n['uuid'] for n in body['nodes']])

        self._associate_node_with_instance()

        _, body = self.client.list_nodes(associated=True)
        self.assertIn(self.node['uuid'], [n['uuid'] for n in body['nodes']])

        _, body = self.client.list_nodes(associated=False)
        self.assertNotIn(self.node['uuid'], [n['uuid'] for n in body['nodes']])

    @decorators.idempotent_id('18c4ebd8-f83a-4df7-9653-9fb33a329730')
    def test_node_port_list(self):
        _, port = self.create_port(self.node['uuid'],
                                   data_utils.rand_mac_address())
        _, body = self.client.list_node_ports(self.node['uuid'])
        self.assertIn(port['uuid'],
                      [p['uuid'] for p in body['ports']])

    @decorators.idempotent_id('72591acb-f215-49db-8395-710d14eb86ab')
    def test_node_port_list_no_ports(self):
        _, node = self.create_node(self.chassis['uuid'])
        _, body = self.client.list_node_ports(node['uuid'])
        self.assertEmpty(body['ports'])

    @decorators.idempotent_id('4fed270a-677a-4d19-be87-fd38ae490320')
    def test_update_node(self):
        props = {'cpu_arch': 'x86_64',
                 'cpus': '12',
                 'local_gb': '10',
                 'memory_mb': '128'}

        _, node = self.create_node(self.chassis['uuid'], **props)

        new_p = {'cpu_arch': 'arm64',
                 'cpus': '1',
                 'local_gb': '10000',
                 'memory_mb': '12300'}

        _, body = self.client.update_node(node['uuid'], properties=new_p)
        _, node = self.client.show_node(node['uuid'])
        self._assertExpected(new_p, node['properties'])

    @decorators.idempotent_id('cbf1f515-5f4b-4e49-945c-86bcaccfeb1d')
    def test_validate_driver_interface(self):
        _, body = self.client.validate_driver_interface(self.node['uuid'])
        core_interfaces = ['power', 'deploy']
        for interface in core_interfaces:
            self.assertIn(interface, body)

    @decorators.idempotent_id('5519371c-26a2-46e9-aa1a-f74226e9d71f')
    def test_set_node_boot_device(self):
        self.client.set_node_boot_device(self.node['uuid'], 'pxe')

    @decorators.idempotent_id('9ea73775-f578-40b9-bc34-efc639c4f21f')
    def test_get_node_boot_device(self):
        body = self.client.get_node_boot_device(self.node['uuid'])
        self.assertIn('boot_device', body)
        self.assertIn('persistent', body)
        self.assertIsInstance(body['boot_device'], str)
        self.assertIsInstance(body['persistent'], bool)

    @decorators.idempotent_id('3622bc6f-3589-4bc2-89f3-50419c66b133')
    def test_get_node_supported_boot_devices(self):
        body = self.client.get_node_supported_boot_devices(self.node['uuid'])
        self.assertIn('supported_boot_devices', body)
        self.assertIsInstance(body['supported_boot_devices'], list)

    @decorators.idempotent_id('f63b6288-1137-4426-8cfe-0d5b7eb87c06')
    def test_get_console(self):
        _, body = self.client.get_console(self.node['uuid'])
        con_info = ['console_enabled', 'console_info']
        for key in con_info:
            self.assertIn(key, body)

    @decorators.idempotent_id('80504575-9b21-4670-92d1-143b948f9437')
    def test_set_console_mode(self):
        self.client.set_console_mode(self.node['uuid'], True)
        waiters.wait_for_bm_node_status(self.client, self.node['uuid'],
                                        'console_enabled', True)

    @decorators.idempotent_id('b02a4f38-5e8b-44b2-aed2-a69a36ecfd69')
    def test_get_node_by_instance_uuid(self):
        instance_uuid = self._associate_node_with_instance()
        _, body = self.client.show_node_by_instance_uuid(instance_uuid)
        self.assertEqual(1, len(body['nodes']))
        self.assertIn(self.node['uuid'], [n['uuid'] for n in body['nodes']])

    @decorators.idempotent_id('b85af8c6-572b-4f20-815e-1cf31844b9f6')
    def test_fault_hidden(self):
        _, loaded_node = self.client.show_node(self.node['uuid'])
        self.assertNotIn('fault', loaded_node)

    @decorators.idempotent_id('e5470656-bb65-4173-be83-2df3fc9aed24')
    def test_conductor_hidden(self):
        _, loaded_node = self.client.show_node(self.node['uuid'])
        self.assertNotIn('conductor', loaded_node)

    @decorators.idempotent_id('5e7f4c54-8216-42d3-83cc-7bd776ffd16f')
    def test_description_hidden(self):
        _, loaded_node = self.client.show_node(self.node['uuid'])
        self.assertNotIn('description', loaded_node)


class TestNodesWithJsonExtSupport(base.BaseBaremetalTest):
    """Tests for baremetal nodes to validate appending .json extension to a

    node's `name` or `uuid` in API versions 1.90 and prior works.
    """

    max_microversion = '1.90'  # Max API version allowed for testing: 1.90
    min_microversion = '1.90'  # Min API version allowed for testing: 1.90

    def setUp(self):
        super(TestNodesWithJsonExtSupport, self).setUp()

        _, self.chassis = self.create_chassis()
        _, self.node = self.create_node(self.chassis['uuid'])

    @decorators.idempotent_id('2ffec2ec-3dca-4e4f-ac6c-dee7c44cdef5')
    def test_update_node_with_json(self):
        """Update a node while appending .json extension to its uuid"""
        props = {'cpu_arch': 'x86_64',
                 'cpus': '12',
                 'local_gb': '10',
                 'memory_mb': '128'}

        _, node = self.create_node(self.chassis['uuid'], **props)

        new_p = {'cpu_arch': 'arm64',
                 'cpus': '1',
                 'local_gb': '10000',
                 'memory_mb': '12300'}

        _, body = self.client.update_node('%s.json' % node['uuid'],
                                          properties=new_p)
        _, node = self.client.show_node(node['uuid'])
        self._assertExpected(new_p, node['properties'])

    @decorators.idempotent_id('4c376a23-04b8-4fc8-955f-abc5668d34c3')
    def test_delete_node_with_json(self):
        """Delete a node while appending .json extension to its uuid"""
        _, node = self.create_node(self.chassis['uuid'])

        self.delete_node('%s.json' % node['uuid'])

        self.assertRaises(lib_exc.NotFound, self.client.show_node,
                          node['uuid'])

    @decorators.idempotent_id('bc1134aa-2950-409c-b91c-b2d85df9b065')
    def test_show_node_with_json(self):
        """Show a node while appending .json extension to its uuid"""
        _, loaded_node = self.client.show_node('%s.json' % self.node['uuid'])
        self._assertExpected(self.node, loaded_node)


class TestNodesWithoutJsonExtSupport(base.BaseBaremetalTest):
    """Tests for baremetal nodes to validate appending .json extension to a

    node's `name` or `uuid` in API versions later than 1.90 is disabled.
    """

    min_microversion = '1.91'  # Min API version allowed for testing: 1.91

    def setUp(self):
        super(TestNodesWithoutJsonExtSupport, self).setUp()

        _, self.chassis = self.create_chassis()
        _, self.node = self.create_node(self.chassis['uuid'])

    @decorators.idempotent_id('65505002-3ac7-4a07-9f66-2e3bcb0f5e95')
    def test_update_node_with_json(self):
        """Trying to update a node while appending .json extension 404s"""
        props = {'cpu_arch': 'x86_64',
                 'cpus': '12',
                 'local_gb': '10',
                 'memory_mb': '128'}

        _, node = self.create_node(self.chassis['uuid'], **props)

        new_p = {'cpu_arch': 'arm64',
                 'cpus': '1',
                 'local_gb': '10000',
                 'memory_mb': '12300'}

        self.assertRaises(lib_exc.NotFound, self.client.update_node,
                          '%s.json' % node['uuid'], properties=new_p)

    @decorators.idempotent_id('594160b6-a608-4024-98d1-818f75d2d895')
    def test_delete_node_with_json(self):
        """Trying to delete a node while appending .json extension 404s"""
        _, node = self.create_node(self.chassis['uuid'])
        self.assertRaises(lib_exc.NotFound, self.client.delete_node,
                          '%s.json' % node['uuid'])

    @decorators.idempotent_id('e83cd539-4d09-46d0-a448-365a219ca353')
    def test_show_node_with_json(self):
        """Trying to show a node while appending .json extension 404s"""

        self.assertRaises(lib_exc.NotFound, self.client.show_node,
                          '%s.json' % self.node['uuid'])


class TestNodesResourceClass(base.BaseBaremetalTest):

    min_microversion = '1.21'

    def setUp(self):
        super(TestNodesResourceClass, self).setUp()
        self.useFixture(
            api_microversion_fixture.APIMicroversionFixture(
                TestNodesResourceClass.min_microversion)
        )
        _, self.chassis = self.create_chassis()
        self.resource_class = data_utils.rand_name(name='Resource_Class')
        _, self.node = self.create_node(
            self.chassis['uuid'], resource_class=self.resource_class)

    @decorators.idempotent_id('2a00340c-8152-4a61-9fc5-0b3cdefec258')
    def test_create_node_resource_class_long(self):
        """Create new node with specified longest name of resource class."""
        res_class_long_name = data_utils.arbitrary_string(80)
        _, body = self.create_node(
            self.chassis['uuid'],
            resource_class=res_class_long_name)
        self.assertEqual(res_class_long_name, body['resource_class'])

    @decorators.idempotent_id('142db00d-ac0f-415b-8da8-9095fbb561f7')
    def test_update_node_resource_class(self):
        """Update existing node with specified resource class."""
        new_res_class_name = data_utils.rand_name(name='Resource_Class')
        _, body = self.client.update_node(
            self.node['uuid'], resource_class=new_res_class_name)
        _, body = self.client.show_node(self.node['uuid'])
        self.assertEqual(new_res_class_name, body['resource_class'])

    @decorators.idempotent_id('73e6f7b5-3e51-49ea-af5b-146cd49f40ee')
    def test_show_node_resource_class(self):
        """Show resource class field of specified node."""
        _, body = self.client.show_node(self.node['uuid'])
        self.assertEqual(self.resource_class, body['resource_class'])

    @decorators.idempotent_id('f2bf4465-280c-4fdc-bbf7-fcf5188befa4')
    def test_list_nodes_resource_class(self):
        """List nodes of specified resource class only."""
        res_class = 'ResClass-{0}'.format(data_utils.rand_uuid())
        for node in range(3):
            _, body = self.create_node(
                self.chassis['uuid'], resource_class=res_class)

        _, body = self.client.list_nodes(resource_class=res_class)
        self.assertEqual(3, len([i['uuid'] for i in body['nodes']]))

    @decorators.idempotent_id('40733bad-bb79-445e-a094-530a44042995')
    def test_list_nodes_detail_resource_class(self):
        """Get detailed nodes list of specified resource class only."""
        res_class = 'ResClass-{0}'.format(data_utils.rand_uuid())
        for node in range(3):
            _, body = self.create_node(
                self.chassis['uuid'], resource_class=res_class)

        _, body = self.client.list_nodes_detail(resource_class=res_class)
        self.assertEqual(3, len([i['uuid'] for i in body['nodes']]))

        for node in body['nodes']:
            self.assertEqual(res_class, node['resource_class'])

    @decorators.attr(type='negative')
    @decorators.idempotent_id('e75136d4-0690-48a5-aef3-75040aee73ad')
    def test_create_node_resource_class_too_long(self):
        """Try to create a node with too long resource class name."""
        resource_class = data_utils.arbitrary_string(81)
        self.assertRaises(lib_exc.BadRequest, self.create_node,
                          self.chassis['uuid'], resource_class=resource_class)

    @decorators.attr(type='negative')
    @decorators.idempotent_id('f0aeece4-8671-44ea-a482-b4047fc4cf74')
    def test_update_node_resource_class_too_long(self):
        """Try to update a node with too long resource class name."""
        resource_class = data_utils.arbitrary_string(81)
        self.assertRaises(lib_exc.BadRequest, self.client.update_node,
                          self.node['uuid'], resource_class=resource_class)


class TestNodesResourceClassOldApi(base.BaseBaremetalTest):

    def setUp(self):
        super(TestNodesResourceClassOldApi, self).setUp()
        _, self.chassis = self.create_chassis()
        _, self.node = self.create_node(self.chassis['uuid'])

    @decorators.attr(type='negative')
    @decorators.idempotent_id('2c364408-4746-4b3c-9821-20d47b57bdec')
    def test_create_node_resource_class_old_api(self):
        """Try to create a node with resource class using older api version."""
        resource_class = data_utils.arbitrary_string()
        self.assertRaises(lib_exc.NotAcceptable, self.create_node,
                          self.chassis['uuid'], resource_class=resource_class)

    @decorators.attr(type='negative')
    @decorators.idempotent_id('666f3c1a-4922-4a3d-b6d9-dea7c74d30bc')
    def test_update_node_resource_class_old_api(self):
        """Try to update a node with resource class using older api version."""
        resource_class = data_utils.arbitrary_string()
        self.assertRaises(lib_exc.NotAcceptable,
                          self.client.update_node,
                          self.node['uuid'], resource_class=resource_class)

    @decorators.attr(type='negative')
    @decorators.idempotent_id('95903480-f16d-4774-8775-6c7f87b27c59')
    def test_list_nodes_by_resource_class_old_api(self):
        """Try to list nodes with resource class using older api version."""
        resource_class = data_utils.arbitrary_string()
        self.assertRaises(
            lib_exc.NotAcceptable,
            self.client.list_nodes, resource_class=resource_class)
        self.assertRaises(
            lib_exc.NotAcceptable,
            self.client.list_nodes_detail, resource_class=resource_class)


class TestNodesVif(base.BaseBaremetalTest):

    min_microversion = '1.28'

    @classmethod
    def skip_checks(cls):
        super(TestNodesVif, cls).skip_checks()
        if not CONF.service_available.neutron:
            raise cls.skipException('Neutron is not enabled.')

    def setUp(self):
        super(TestNodesVif, self).setUp()

        if CONF.network.shared_physical_network:
            if not CONF.compute.fixed_network_name:
                raise self.skipException(
                    'Unable to perform vif test as there is no network '
                    'defined for the [compute]fixed_network_name '
                    'parameter.')
            self.net = self.os_admin.networks_client.list_networks(
                name=CONF.compute.fixed_network_name)['networks'][0]
        else:
            self.net = self.os_admin.networks_client.\
                create_network()['network']
            self.addCleanup(self.os_admin.networks_client.delete_network,
                            self.net['id'])

        _, self.chassis = self.create_chassis()
        # The tests will mostly fail in this class if exposed to the
        # noop network interface, which is what the default is.
        _, self.node = self.create_node(self.chassis['uuid'],
                                        network_interface='flat')

        self.nport_id = self.os_admin.ports_client.create_port(
            network_id=self.net['id'])['port']['id']
        self.addCleanup(self.os_admin.ports_client.delete_port,
                        self.nport_id)

    @decorators.idempotent_id('a3d319d0-cacb-4e55-a3dc-3fa8b74880f1')
    def test_vif_on_port(self):
        """Test attachment and detachment of VIFs on the node with port.

        Test steps:
        1) Create chassis and node in setUp.
        2) Create port for the node.
        3) Attach VIF to the node.
        4) Check VIF info in VIFs list and port internal_info.
        5) Detach VIF from the node.
        6) Check that no more VIF info in VIFs list and port internal_info.
        """
        self.useFixture(
            api_microversion_fixture.APIMicroversionFixture('1.28'))
        _, self.port = self.create_port(self.node['uuid'],
                                        data_utils.rand_mac_address())
        self.client.vif_attach(self.node['uuid'], self.nport_id)
        _, body = self.client.vif_list(self.node['uuid'])
        self.assertEqual({'vifs': [{'id': self.nport_id}]}, body)
        _, port = self.client.show_port(self.port['uuid'])
        self.assertEqual(self.nport_id,
                         port['internal_info']['tenant_vif_port_id'])
        self.client.vif_detach(self.node['uuid'], self.nport_id)
        _, body = self.client.vif_list(self.node['uuid'])
        self.assertEqual({'vifs': []}, body)
        _, port = self.client.show_port(self.port['uuid'])
        self.assertNotIn('tenant_vif_port_id', port['internal_info'])

    @decorators.idempotent_id('95279515-7d0a-4f5f-987f-93e36aae5585')
    def test_vif_on_portgroup(self):
        """Test attachment and detachment of VIFs on the node with port group.

        Test steps:
        1) Create chassis and node in setUp.
        2) Create port for the node.
        3) Create port group for the node.
        4) Plug port into port group.
        5) Attach VIF to the node.
        6) Check VIF info in VIFs list and port group internal_info, but
           not in port internal_info.
        7) Detach VIF from the node.
        8) Check that no VIF info in VIFs list and port group internal_info.
        """
        self.useFixture(
            api_microversion_fixture.APIMicroversionFixture('1.28'))
        _, self.port = self.create_port(self.node['uuid'],
                                        data_utils.rand_mac_address())
        _, self.portgroup = self.create_portgroup(
            self.node['uuid'], address=data_utils.rand_mac_address())

        patch = [{'path': '/portgroup_uuid',
                  'op': 'add',
                  'value': self.portgroup['uuid']}]
        self.client.update_port(self.port['uuid'], patch)

        self.client.vif_attach(self.node['uuid'], self.nport_id)
        _, body = self.client.vif_list(self.node['uuid'])
        self.assertEqual({'vifs': [{'id': self.nport_id}]}, body)

        _, port = self.client.show_port(self.port['uuid'])
        self.assertNotIn('tenant_vif_port_id', port['internal_info'])
        _, portgroup = self.client.show_portgroup(self.portgroup['uuid'])
        self.assertEqual(self.nport_id,
                         portgroup['internal_info']['tenant_vif_port_id'])

        self.client.vif_detach(self.node['uuid'], self.nport_id)
        _, body = self.client.vif_list(self.node['uuid'])
        self.assertEqual({'vifs': []}, body)
        _, portgroup = self.client.show_portgroup(self.portgroup['uuid'])
        self.assertNotIn('tenant_vif_port_id', portgroup['internal_info'])

    @decorators.idempotent_id('a3d319d0-cacb-4e55-a3dc-3fa8b74880f2')
    def test_vif_already_set(self):
        self.useFixture(
            api_microversion_fixture.APIMicroversionFixture('1.28'))
        _, self.port = self.create_port(self.node['uuid'],
                                        data_utils.rand_mac_address())
        self.client.vif_attach(self.node['uuid'], self.nport_id)
        _, body = self.client.vif_list(self.node['uuid'])
        self.assertEqual({'vifs': [{'id': self.nport_id}]}, body)

        self.assertRaises(lib_exc.Conflict, self.client.vif_attach,
                          self.node['uuid'], self.nport_id)

        self.client.vif_detach(self.node['uuid'], self.nport_id)

    @decorators.attr(type='negative')
    @decorators.idempotent_id('628350f8-4498-4204-b546-f3c01b93c7e3')
    def test_vif_already_attached_on_internal_info(self):
        """Negative test for duplicated attachment/detachment of VIFs.

        Test steps:
        1) Create chassis and node in setUp.
        2) Create port for the node.
        3) Attach VIF to the node.
        4) Try to attach the same VIF to the node.
        5) Detach VIF from the node.
        6) Try to detach the same VIF from the node.
        """
        self.useFixture(
            api_microversion_fixture.APIMicroversionFixture('1.28'))
        _, self.port = self.create_port(self.node['uuid'],
                                        data_utils.rand_mac_address())
        self.client.vif_attach(self.node['uuid'], self.nport_id)
        _, body = self.client.vif_list(self.node['uuid'])

        self.assertEqual({'vifs': [{'id': self.nport_id}]}, body)

        self.assertRaises(lib_exc.Conflict, self.client.vif_attach,
                          self.node['uuid'], self.nport_id)
        self.client.vif_detach(self.node['uuid'], self.nport_id)
        self.assertRaises(lib_exc.BadRequest, self.client.vif_detach,
                          self.node['uuid'], self.nport_id)

    @decorators.attr(type='negative')
    @decorators.idempotent_id('ec0c14a4-6853-4907-9091-755c9c6c152f')
    def test_vif_already_attached_with_portgroups(self):
        """Negative test: try duplicated attachment of VIFs with port groups.

        Test steps:
        1) Create chassis and node in setUp.
        2) Create port for the node.
        3) Create port group for the node.
        4) Plug port into port group.
        5) Attach VIF to the node.
        6) Try to attach the same VIF to the node.
        7) Check that VIF was not attached to port when portgroup is busy.
        8) Detach VIF from the node.
        9) Check there is no VIF data in port group internal info.
        """
        _, self.port = self.create_port(self.node['uuid'],
                                        data_utils.rand_mac_address())
        _, self.portgroup = self.create_portgroup(
            self.node['uuid'], address=data_utils.rand_mac_address())

        patch = [{'path': '/portgroup_uuid',
                  'op': 'add',
                  'value': self.portgroup['uuid']}]
        self.client.update_port(self.port['uuid'], patch)
        self.client.vif_attach(self.node['uuid'], self.nport_id)
        self.assertRaises(lib_exc.Conflict, self.client.vif_attach,
                          self.node['uuid'], self.nport_id)

        _, port = self.client.show_port(self.port['uuid'])
        self.assertNotIn('tenant_vif_port_id', port['internal_info'])

        self.client.vif_detach(self.node['uuid'], self.nport_id)
        _, portgroup = self.client.show_portgroup(self.portgroup['uuid'])
        self.assertNotIn('tenant_vif_port_id', portgroup['internal_info'])

    @decorators.attr(type='negative')
    @decorators.idempotent_id('91e08d6a-0438-4171-b404-bc86b0bc8861')
    def test_vif_attach_no_free_port(self):
        """Negative test for VIF attachment attempt with no free ports.

        Test steps:
        1) Create chassis and node in setUp.
        2) Create port for the node.
        3) Attach VIF to the node.
        4) Try to attach new VIF to the same node with port.
        5) Check that VIF still attached with original port.
        """
        _, self.port = self.create_port(self.node['uuid'],
                                        data_utils.rand_mac_address())
        self.client.vif_attach(self.node['uuid'], self.nport_id)
        self.assertRaises(lib_exc.BadRequest, self.client.vif_attach,
                          self.node['uuid'], 'test-vif-new')
        _, port = self.client.show_port(self.port['uuid'])
        self.assertEqual(self.nport_id,
                         port['internal_info']['tenant_vif_port_id'])
        self.client.vif_detach(self.node['uuid'], self.nport_id)

    @decorators.attr(type='negative')
    @decorators.idempotent_id('85b610cd-5ba8-49a7-8ce2-5e364056fd29')
    def test_vif_attach_no_port(self):
        """Negative test for VIF attachment attempt with no ports."""
        self.assertRaises(lib_exc.BadRequest,
                          self.client.vif_attach,
                          self.node['uuid'], self.nport_id)

    @decorators.attr(type='negative')
    @decorators.idempotent_id('481777b8-8f0d-4932-af3c-1a8e166949f3')
    def test_vif_attach_with_empty_portgroup(self):
        """Negative test for VIF attachment attempt to empty port group.

        Test steps:
        1) Create chassis and node in setUp.
        2) Create a port group for the node.
        3) Try to attach VIF to the node.
        4) Check that VIF info did not get to port group when there is no port.
        """
        _, self.portgroup = self.create_portgroup(
            self.node['uuid'], address=data_utils.rand_mac_address())
        self.assertRaises(lib_exc.BadRequest, self.client.vif_attach,
                          self.node['uuid'], self.nport_id)
        _, portgroup = self.client.show_portgroup(self.portgroup['uuid'])
        self.assertNotIn('tenant_vif_port_id', portgroup['internal_info'])

    @decorators.attr(type='negative')
    @decorators.idempotent_id('5572fecc-bc22-461a-a891-7f74e2e936bd')
    def test_vif_attach_port_not_in_portgroup(self):
        """Negative test for VIF attachment to the port not in port group.

        Test steps:
        1) Create chassis and node in setUp.
        2) Create a port group for the node.
        3) Create a port for the node.
        4) Attach VIF to the node.
        5) Check that VIF was attached to the port.
        6) Check that VIF was NOT attached to the portgroup.
        """
        _, self.portgroup = self.create_portgroup(
            self.node['uuid'], address=data_utils.rand_mac_address())
        _, self.port = self.create_port(self.node['uuid'],
                                        data_utils.rand_mac_address())
        self.client.vif_attach(self.node['uuid'], self.nport_id)
        _, port = self.client.show_port(self.port['uuid'])
        self.assertEqual(self.nport_id,
                         port['internal_info']['tenant_vif_port_id'])
        _, portgroup = self.client.show_portgroup(self.portgroup['uuid'])
        self.assertNotIn('tenant_vif_port_id', portgroup['internal_info'])
        self.client.vif_detach(self.node['uuid'], self.nport_id)

    @decorators.attr(type='negative')
    @decorators.idempotent_id('3affca81-9f3f-4dab-ad3d-77c892d8d0d7')
    def test_vif_attach_node_doesnt_exist(self):
        """Negative test to try to attach VIF to not-existing node."""
        self.assertRaises(lib_exc.NotFound,
                          self.client.vif_attach,
                          data_utils.rand_uuid(),
                          self.nport_id)

    @decorators.attr(type='negative')
    @decorators.idempotent_id('da036225-47b0-43b7-9586-0d6390bd3cd9')
    def test_vif_detach_not_existing(self):
        """Negative test for VIF detachment of not existing VIF."""
        self.assertRaises(lib_exc.BadRequest,
                          self.client.vif_detach,
                          self.node['uuid'], self.nport_id)

    @decorators.attr(type='negative')
    @decorators.idempotent_id('ff3c9ce2-4979-4a54-a860-fab088a6669f')
    def test_vif_detach_no_args(self):
        """Negative test for VIF detachment with lack of arguments."""
        self.assertRaises(lib_exc.BadRequest,
                          self.client.vif_detach,
                          self.node['uuid'], '')
        self.assertRaises(lib_exc.NotFound,
                          self.client.vif_detach,
                          '', '')
        self.assertRaises(lib_exc.NotFound,
                          self.client.vif_detach,
                          '', self.nport_id)


class TestHardwareInterfaces(base.BaseBaremetalTest):

    min_microversion = '1.31'
    # Subclasses can override this with more interfaces available in later API
    # versions.
    hardware_interfaces = [
        'boot',
        'console',
        'deploy',
        'inspect',
        'management',
        'power',
        'raid',
        'vendor',
    ]

    @classmethod
    def skip_checks(cls):
        super(TestHardwareInterfaces, cls).skip_checks()
        if CONF.baremetal.driver != 'fake-hardware':
            raise cls.skipException('These tests rely on fake-hardware')

    @property
    def optional_interfaces(self):
        return set(self.hardware_interfaces) - {'boot', 'deploy',
                                                'management', 'power'}

    def setUp(self):
        super(TestHardwareInterfaces, self).setUp()

        _, self.chassis = self.create_chassis()
        _, self.node = self.create_node(self.chassis['uuid'])

        # Reset optional interfaces to non-default values
        for iface in self.optional_interfaces:
            self.client.update_node(self.node['uuid'],
                                    [{'path': '/%s_interface' % iface,
                                      'op': 'add',
                                      'value': 'no-%s' % iface}])

    def test_set_interfaces(self):
        for iface in self.hardware_interfaces:
            field = '%s_interface' % iface
            self.client.update_node(self.node['uuid'],
                                    [{'path': '/%s' % field,
                                      'op': 'add',
                                      'value': 'fake'}])
            _, node = self.client.show_node(self.node['uuid'])
            self.assertEqual('fake', node[field])

    def test_reset_interfaces(self):
        for iface in self.hardware_interfaces:
            field = '%s_interface' % iface
            self.client.update_node(self.node['uuid'],
                                    [{'path': '/%s' % field,
                                      'op': 'remove'}])
            _, node = self.client.show_node(self.node['uuid'])
            if (iface in ['boot', 'deploy', 'inspect', 'network']
                    and 'fake' != node[field]):
                # NOTE(TheJulia): Depending on the remote configuration
                # these interfaces may have explicit defaults which the
                # reset action changes the value to. This is okay, but
                # we need to not fail when that is the case.
                self.assertNotEqual(self.node[field], node[field])
            else:
                self.assertEqual('fake', node[field])


class TestResetInterfaces(TestHardwareInterfaces):

    min_microversion = '1.45'

    @classmethod
    def skip_checks(cls):
        super(TestResetInterfaces, cls).skip_checks()
        if 'ipmi' not in CONF.baremetal.enabled_hardware_types:
            raise cls.skipException('These tests rely on ipmi enabled')

    def test_no_reset_by_default(self):
        self.assertRaises(
            lib_exc.BadRequest,
            self.client.update_node,
            self.node['uuid'],
            [{'path': '/driver', 'value': 'ipmi', 'op': 'replace'}])
        _, node = self.client.show_node(self.node['uuid'])
        self.assertEqual('fake-hardware', node['driver'])

    def test_reset_all_interfaces(self):
        self.client.update_node(self.node['uuid'],
                                [{'path': '/driver',
                                  'value': 'ipmi',
                                  'op': 'replace'}],
                                reset_interfaces=True)
        _, node = self.client.show_node(self.node['uuid'])
        for iface in self.hardware_interfaces:
            self.assertNotEqual('fake', node['%s_interface' % iface])


class TestNodesTraits(base.BaseBaremetalTest):

    min_microversion = '1.37'

    def setUp(self):
        super(TestNodesTraits, self).setUp()
        self.useFixture(
            api_microversion_fixture.APIMicroversionFixture(
                TestNodesTraits.min_microversion)
        )
        _, self.chassis = self.create_chassis()
        # One standard trait & one custom trait.
        self.traits = ['CUSTOM_TRAIT1', 'HW_CPU_X86_VMX']
        _, self.node = self.create_node(self.chassis['uuid'])

    @decorators.idempotent_id('5c3a2dd0-af10-474d-a209-d30426e1eb5d')
    def test_list_node_traits(self):
        """List traits for a node."""
        _, body = self.client.list_node_traits(self.node['uuid'])
        self.assertEqual([], body['traits'])

        self.client.set_node_traits(self.node['uuid'], self.traits)
        _, body = self.client.list_node_traits(self.node['uuid'])
        self.assertEqual(self.traits, sorted(body['traits']))

    @decorators.attr(type='negative')
    @decorators.idempotent_id('3b83dbd3-4a89-4173-920a-ca33ed3aad69')
    def test_list_node_traits_non_existent_node(self):
        """Try to list traits for a non-existent node."""
        node_uuid = uuidutils.generate_uuid()
        self.assertRaises(
            lib_exc.NotFound,
            self.client.list_node_traits, node_uuid)

    @decorators.idempotent_id('aa961bf6-ea2f-484b-961b-eae2da0e6b7e')
    def test_set_node_traits(self):
        """Set the traits for a node."""
        self.client.set_node_traits(self.node['uuid'], self.traits)

        _, body = self.client.list_node_traits(self.node['uuid'])
        self.assertEqual(self.traits, sorted(body['traits']))

        self.client.set_node_traits(self.node['uuid'], [])

        _, body = self.client.list_node_traits(self.node['uuid'])
        self.assertEqual([], body['traits'])

    @decorators.idempotent_id('727a5e11-5654-459f-8af6-e14eb987a283')
    def test_set_node_traits_max_traits(self):
        """Set the maximum number of traits for a node."""
        traits = ['CUSTOM_TRAIT%d' % i for i in range(50)]
        self.client.set_node_traits(self.node['uuid'], traits)

        _, body = self.client.list_node_traits(self.node['uuid'])
        self.assertEqual(sorted(traits), sorted(body['traits']))

    @decorators.attr(type='negative')
    @decorators.idempotent_id('75831f5d-ca44-403b-8fd6-f7cad95b1c54')
    def test_set_node_traits_too_many(self):
        """Set more than the maximum number of traits for a node."""
        traits = ['CUSTOM_TRAIT%d' % i for i in range(51)]
        self.assertRaises(
            lib_exc.BadRequest,
            self.client.set_node_traits, self.node['uuid'], traits)

        _, body = self.client.list_node_traits(self.node['uuid'])
        self.assertEqual([], body['traits'])

    @decorators.idempotent_id('d81ceeab-a50f-427a-bc5a-aa916478d0d3')
    def test_set_node_traits_duplicate_trait(self):
        """Set the traits for a node, ensuring duplicates are ignored."""
        self.client.set_node_traits(self.node['uuid'],
                                    ['CUSTOM_TRAIT1', 'CUSTOM_TRAIT1'])

        _, body = self.client.list_node_traits(self.node['uuid'])
        self.assertEqual(['CUSTOM_TRAIT1'], body['traits'])

    @decorators.attr(type='negative')
    @decorators.idempotent_id('2fb4c9d9-8e5b-4189-b547-26596014491c')
    def test_set_node_traits_non_existent_node(self):
        """Try to set traits for a non-existent node."""
        node_uuid = uuidutils.generate_uuid()
        self.assertRaises(
            lib_exc.NotFound,
            self.client.set_node_traits, node_uuid, ['CUSTOM_TRAIT1'])

    @decorators.idempotent_id('47db09d9-af2b-424d-9d51-7efca2920f20')
    def test_add_node_trait_long(self):
        """Add a node trait of the largest possible length."""
        trait_long_name = 'CUSTOM_' + data_utils.arbitrary_string(248).upper()
        self.client.add_node_trait(self.node['uuid'], trait_long_name)

        _, body = self.client.list_node_traits(self.node['uuid'])
        self.assertEqual([trait_long_name], body['traits'])

    @decorators.attr(type='negative')
    @decorators.idempotent_id('2a4daa8d-2b85-40ac-a8a0-0462cc9a57ef')
    def test_add_node_trait_too_long(self):
        """Try to add a node trait longer than the largest possible length."""
        trait_long_name = 'CUSTOM_' + data_utils.arbitrary_string(249).upper()
        self.assertRaises(
            lib_exc.BadRequest,
            self.client.add_node_trait, self.node['uuid'], trait_long_name)

        _, body = self.client.list_node_traits(self.node['uuid'])
        self.assertEqual([], body['traits'])

    @decorators.idempotent_id('4b737e7f-101e-493e-b5ce-494fbffe18fd')
    def test_add_node_trait_duplicate_trait(self):
        """Add a node trait that already exists."""
        self.client.add_node_trait(self.node['uuid'], 'CUSTOM_TRAIT1')
        self.client.add_node_trait(self.node['uuid'], 'CUSTOM_TRAIT1')

        _, body = self.client.list_node_traits(self.node['uuid'])
        self.assertEqual(['CUSTOM_TRAIT1'], body['traits'])

    @decorators.attr(type='negative')
    @decorators.idempotent_id('65bce181-89ce-435e-a7d8-3ba60aafd08d')
    def test_add_node_trait_too_many(self):
        """Add a trait to a node that would exceed the maximum."""
        traits = ['CUSTOM_TRAIT%d' % i for i in range(50)]
        self.client.set_node_traits(self.node['uuid'], traits)

        self.assertRaises(
            lib_exc.BadRequest,
            self.client.add_node_trait, self.node['uuid'], 'CUSTOM_TRAIT50')

        _, body = self.client.list_node_traits(self.node['uuid'])
        self.assertEqual(sorted(traits), sorted(body['traits']))

    @decorators.attr(type='negative')
    @decorators.idempotent_id('cca0e831-32af-4ce9-bfce-d3834fea57aa')
    def test_add_node_trait_non_existent_node(self):
        """Try to add a trait to a non-existent node."""
        node_uuid = uuidutils.generate_uuid()
        self.assertRaises(
            lib_exc.NotFound,
            self.client.add_node_trait, node_uuid, 'CUSTOM_TRAIT1')

    @decorators.idempotent_id('e4bf8bf0-3004-44bc-8bfe-f9f1a167d999')
    def test_remove_node_traits(self):
        """Remove all traits from a node."""
        self.client.set_node_traits(self.node['uuid'], self.traits)

        self.client.remove_node_traits(self.node['uuid'])

        _, body = self.client.list_node_traits(self.node['uuid'])
        self.assertEqual([], body['traits'])

    @decorators.idempotent_id('4d8c9a35-0036-4139-85c1-5f242395680f')
    def test_remove_node_traits_no_traits(self):
        """Remove all traits from a node that has no traits."""
        self.client.remove_node_traits(self.node['uuid'])

        _, body = self.client.list_node_traits(self.node['uuid'])
        self.assertEqual([], body['traits'])

    @decorators.attr(type='negative')
    @decorators.idempotent_id('625c911a-48e8-4bef-810b-7cf33c0846a2')
    def test_remove_node_traits_non_existent_node(self):
        """Try to remove all traits from a non-existent node."""
        node_uuid = uuidutils.generate_uuid()
        self.assertRaises(
            lib_exc.NotFound,
            self.client.remove_node_traits, node_uuid)

    @decorators.idempotent_id('3591d514-39b9-425e-9afe-ea74ae347486')
    def test_remove_node_trait(self):
        """Remove a trait from a node."""
        self.client.set_node_traits(self.node['uuid'], self.traits)

        self.client.remove_node_trait(self.node['uuid'], 'CUSTOM_TRAIT1')

        _, body = self.client.list_node_traits(self.node['uuid'])
        self.assertEqual(['HW_CPU_X86_VMX'], body['traits'])

    @decorators.attr(type='negative')
    @decorators.idempotent_id('b50ae543-5e5e-4b1a-b2f2-9e00fe55974b')
    def test_remove_node_trait_non_existent_trait(self):
        """Try to remove a non-existent trait from a node."""
        self.assertRaises(
            lib_exc.NotFound,
            self.client.remove_node_trait, self.node['uuid'], 'CUSTOM_TRAIT1')

    @decorators.attr(type='negative')
    @decorators.idempotent_id('f1469745-7cdf-4cae-9699-73d029c47bc3')
    def test_remove_node_trait_non_existent_node(self):
        """Try to remove a trait from a non-existent node."""
        node_uuid = uuidutils.generate_uuid()
        self.assertRaises(
            lib_exc.NotFound,
            self.client.remove_node_trait, node_uuid, 'CUSTOM_TRAIT1')

    @decorators.idempotent_id('03f9e57f-e584-448a-926f-53035e583e7e')
    def test_list_nodes_detail(self):
        """Get detailed nodes list."""
        self.client.set_node_traits(self.node['uuid'], self.traits)

        _, body = self.client.list_nodes_detail()
        self.assertGreaterEqual(len(body['nodes']), 1)

        for node in body['nodes']:
            self.assertIn('traits', node)
            if node['uuid'] == self.node['uuid']:
                self.assertEqual(self.traits, sorted(node['traits']))

    @decorators.idempotent_id('2b82f704-1580-403a-af92-92c29a7eebb7')
    def test_list_nodes_traits_field(self):
        """Get nodes list with the traits field."""
        self.client.set_node_traits(self.node['uuid'], self.traits)

        _, body = self.client.list_nodes(fields='uuid,traits')
        self.assertGreaterEqual(len(body['nodes']), 1)

        for node in body['nodes']:
            self.assertIn('traits', node)
            if node['uuid'] == self.node['uuid']:
                self.assertEqual(self.traits, sorted(node['traits']))

    @decorators.idempotent_id('c83c537a-76aa-4d8a-8673-128d01ee403d')
    def test_show_node(self):
        """Show a node with traits."""
        self.client.set_node_traits(self.node['uuid'], self.traits)

        _, body = self.client.show_node(self.node['uuid'])

        self.assertIn('traits', body)
        self.assertEqual(self.traits, sorted(body['traits']))

    @decorators.attr(type='negative')
    @decorators.idempotent_id('9ab6a19c-83b9-4600-b55b-325a51e2f8f6')
    def test_update_node_traits(self):
        """Try updating an existing node with traits."""
        patch = [{'path': '/traits',
                  'op': 'add',
                  'value': ['CUSTOM_TRAIT1']}]
        self.assertRaises(
            lib_exc.BadRequest,
            self.client.update_node, self.node['uuid'], patch)

        _, body = self.client.list_node_traits(self.node['uuid'])
        self.assertEqual([], body['traits'])


class TestNodesTraitsOldApi(base.BaseBaremetalTest):

    def setUp(self):
        super(TestNodesTraitsOldApi, self).setUp()
        _, self.chassis = self.create_chassis()
        _, self.node = self.create_node(self.chassis['uuid'])

    @decorators.attr(type='negative')
    @decorators.idempotent_id('5419af7b-4e27-4be4-88f6-e01c598a8102')
    def test_list_node_traits_old_api(self):
        """Try to list traits for a node using an older api version."""
        self.assertRaises(
            lib_exc.NotAcceptable,
            self.client.list_node_traits, self.node['uuid'])

    @decorators.attr(type='negative')
    @decorators.idempotent_id('a4353f3a-bedc-4579-9c7e-4bebcd95903d')
    def test_add_node_trait_old_api(self):
        """Try to add a trait to a node using an older api version."""
        exc = self.assertRaises(
            lib_exc.UnexpectedResponseCode,
            self.client.add_node_trait, self.node['uuid'], 'CUSTOM_TRAIT1')
        self.assertEqual(405, exc.resp.status)

    @decorators.attr(type='negative')
    @decorators.idempotent_id('91cc43d8-2f6f-4b1b-95e9-68dedca54e6b')
    def test_set_node_traits_old_api(self):
        """Try to set traits for a node using an older api version."""
        exc = self.assertRaises(
            lib_exc.UnexpectedResponseCode,
            self.client.set_node_traits, self.node['uuid'], ['CUSTOM_TRAIT1'])
        self.assertEqual(405, exc.resp.status)

    @decorators.attr(type='negative')
    @decorators.idempotent_id('0f9af890-a57a-4c25-86c8-6418d1b8f4d4')
    def test_remove_node_trait_old_api(self):
        """Try to remove a trait from a node using an older api version."""
        self.assertRaises(
            lib_exc.NotFound,
            self.client.remove_node_trait, self.node['uuid'], 'CUSTOM_TRAIT1')

    @decorators.attr(type='negative')
    @decorators.idempotent_id('f8375b3c-1939-4d1c-97c4-d23e10680090')
    def test_remove_node_traits_old_api(self):
        """Try to remove all traits from a node using an older api version."""
        self.assertRaises(
            lib_exc.NotFound,
            self.client.remove_node_traits, self.node['uuid'])

    @decorators.attr(type='negative')
    @decorators.idempotent_id('525eeb59-b7ce-413d-a37b-401e67402a4c')
    def test_list_nodes_detail_old_api(self):
        """Get detailed nodes list, ensure they have no traits."""
        _, body = self.client.list_nodes_detail()
        self.assertGreaterEqual(len(body['nodes']), 1)

        for node in body['nodes']:
            self.assertNotIn('traits', node)

    @decorators.attr(type='negative')
    @decorators.idempotent_id('eb75b3c8-ac9c-4399-90a2-c0030bfde7a6')
    def test_list_nodes_traits_field(self):
        """Try to list nodes' traits field using older api version."""
        self.assertRaises(
            lib_exc.NotAcceptable,
            self.client.list_nodes, fields='traits')

    @decorators.attr(type='negative')
    @decorators.idempotent_id('214ae7fc-149b-4657-b6bc-66353d49ade8')
    def test_show_node_old_api(self):
        """Show a node, ensure it has no traits."""
        _, body = self.client.show_node(self.node['uuid'])
        self.assertNotIn('traits', body)


class TestNodeFault(base.BaseBaremetalTest):
    """Tests for fault of baremetal nodes."""

    min_microversion = '1.42'

    def setUp(self):
        super(TestNodeFault, self).setUp()

        _, self.chassis = self.create_chassis()
        _, self.node = self.create_node(self.chassis['uuid'])

    @decorators.idempotent_id('649b4660-4f76-4d67-94df-6631a2cb2cd9')
    def test_fault_shown(self):
        _, loaded_node = self.client.show_node(self.node['uuid'])
        self.assertIn('fault', loaded_node)

    @decorators.idempotent_id('62f453be-8f30-4cfe-a19a-23656068e546')
    def test_list_nodes_fault(self):
        _, body = self.client.list_nodes()
        self.assertIn(self.node['uuid'], [n['uuid'] for n in body['nodes']])

        _, body = self.client.list_nodes(fault='power failure')
        self.assertNotIn(self.node['uuid'],
                         [n['uuid'] for n in body['nodes']])

    @decorators.idempotent_id('c8fb55f1-873f-4fb9-bd57-6f1de0479873')
    def test_list_nodes_with_invalid_fault(self):
        self.assertRaises(
            lib_exc.BadRequest,
            self.client.list_nodes, fault='somefake')


class TestNodeProtected(base.BaseBaremetalTest):
    """Tests for protected baremetal nodes."""

    min_microversion = '1.48'

    def setUp(self):
        super(TestNodeProtected, self).setUp()

        _, self.chassis = self.create_chassis()
        _, self.node = self.create_node(
            self.chassis['uuid'],
            # Fake deploy interface to bypass cleaning in the test flow.
            deploy_interface='fake',
            # Noop network interface to skip networking involvement.
            network_interface='noop')
        self.provide_node(self.node['uuid'])

    @decorators.idempotent_id('52f0cb1c-ad7b-43dc-8e22-a76438b67716')
    def test_node_protected_set_unset(self):
        self.deploy_node(self.node['uuid'])
        _, self.node = self.client.show_node(self.node['uuid'])
        self.assertFalse(self.node['protected'])
        self.assertIsNone(self.node['protected_reason'])

        self.client.update_node(self.node['uuid'], protected=True,
                                protected_reason='reason!')
        _, self.node = self.client.show_node(self.node['uuid'])
        self.assertTrue(self.node['protected'])
        self.assertEqual('reason!', self.node['protected_reason'])

        self.client.update_node(self.node['uuid'], protected=False)
        _, self.node = self.client.show_node(self.node['uuid'])
        self.assertFalse(self.node['protected'])
        self.assertIsNone(self.node['protected_reason'])

    @decorators.idempotent_id('8fbd101e-90e6-4843-b41a-556b34802972')
    def test_node_protected(self):
        self.deploy_node(self.node['uuid'])
        self.client.update_node(self.node['uuid'], protected=True)

        self.assertRaises(lib_exc.Forbidden,
                          self.set_node_provision_state,
                          self.node['uuid'], 'deleted', 'available')
        self.assertRaises(lib_exc.Forbidden,
                          self.set_node_provision_state,
                          self.node['uuid'], 'rebuild', 'active')
        self.client.update_node(self.node['uuid'], protected=False)

    @decorators.idempotent_id('04a21b51-2991-4213-8c2f-a96cfdada802')
    def test_node_protected_from_deletion(self):
        self.deploy_node(self.node['uuid'])
        self.client.update_node(self.node['uuid'], protected=True,
                                maintenance=True)
        # undo maintenance because we can't teardown something
        # in maintenance.
        self.addCleanup(self.client.update_node, self.node['uuid'],
                        maintenance=False)
        self.assertRaises(lib_exc.Forbidden,
                          self.client.delete_node,
                          self.node['uuid'])
        self.client.update_node(self.node['uuid'], protected=False)

    @decorators.attr(type='negative')
    @decorators.idempotent_id('1c819f4c-6c1d-4150-ba4a-3b0dcb3c8694')
    def test_node_protected_negative(self):
        # Cannot be set for available nodes
        self.assertRaises(lib_exc.Conflict,
                          self.client.update_node,
                          self.node['uuid'], protected=True)

        self.deploy_node(self.node['uuid'])

        # Reason cannot be set for nodes that are not protected
        self.assertRaises(lib_exc.BadRequest,
                          self.client.update_node,
                          self.node['uuid'], protected_reason='reason!')
        self.client.update_node(self.node['uuid'], protected=False)


class TestNodesProtectedOldApi(base.BaseBaremetalTest):

    def setUp(self):
        super(TestNodesProtectedOldApi, self).setUp()
        _, self.chassis = self.create_chassis()

        _, self.node = self.create_node(self.chassis['uuid'])
        self.useFixture(
            api_microversion_fixture.APIMicroversionFixture('1.31'))
        # Now with a 1.31 microversion, swap the deploy and network
        # interfaces into place so the test doesn't break depending on
        # the environment's default state.
        self.client.update_node(self.node['uuid'],
                                [{'path': '/deploy_interface',
                                  'op': 'replace',
                                  'value': 'fake'},
                                 {'path': '/network_interface',
                                  'op': 'replace',
                                  'value': 'noop'}])
        self.useFixture(api_microversion_fixture.APIMicroversionFixture('1.1'))
        self.deploy_node(self.node['uuid'])
        _, self.node = self.client.show_node(self.node['uuid'])

    @decorators.attr(type='negative')
    @decorators.idempotent_id('08971546-27cc-40ab-851e-ba7bb52c00ab')
    def test_node_protected_old_api(self):
        exc = self.assertRaises(
            lib_exc.RestClientException,
            self.client.update_node, self.node['uuid'], protected=True)
        # 400 for old ironic, 406 for new ironic with old microversion.
        self.assertIn(exc.resp.status, (400, 406))


class TestNodeConductor(base.BaseBaremetalTest):
    """Tests for conductor field of baremetal nodes."""

    min_microversion = '1.49'

    def setUp(self):
        super(TestNodeConductor, self).setUp()

        _, self.chassis = self.create_chassis()
        _, self.node = self.create_node(self.chassis['uuid'])

    @decorators.idempotent_id('1af888b2-2a19-43da-8181-a5381d6ff536')
    def test_conductor_exposed(self):
        _, loaded_node = self.client.show_node(self.node['uuid'])
        self.assertIn('conductor', loaded_node)

    @decorators.idempotent_id('53bcef99-2989-4755-aa8f-c31037cd15de')
    def test_list_nodes_by_conductor(self):
        _, loaded_node = self.client.show_node(self.node['uuid'])
        hostname = loaded_node['conductor']

        _, nodes = self.client.list_nodes(conductor=hostname)
        self.assertIn(self.node['uuid'],
                      [n['uuid'] for n in nodes['nodes']])


class TestNodeDescription(base.BaseBaremetalTest):
    """Tests for the description field."""

    min_microversion = '1.51'

    def setUp(self):
        super(TestNodeDescription, self).setUp()

        _, self.chassis = self.create_chassis()
        _, self.node = self.create_node(self.chassis['uuid'])

    @decorators.idempotent_id('66d0da49-e5ac-4f49-b065-9d2207d8a3af')
    def test_description_exposed(self):
        _, loaded_node = self.client.show_node(self.node['uuid'])
        self.assertIn('description', loaded_node)

    @decorators.idempotent_id('85b4a4b5-37e5-4b60-8dc7-f5a26dfa78a3')
    def test_node_description_set_unset(self):
        self.client.update_node(self.node['uuid'], description='meow')
        _, self.node = self.client.show_node(self.node['uuid'])
        self.assertEqual('meow', self.node['description'])

        self.client.update_node(self.node['uuid'], description=None)
        _, self.node = self.client.show_node(self.node['uuid'])
        self.assertIsNone(self.node['description'])

    @decorators.idempotent_id('3d649bb3-a58b-4b9e-8dfa-41ab634b1153')
    def test_create_node_with_description(self):
        _, body = self.create_node(self.chassis['uuid'], description='meow')
        self.assertEqual('meow', body['description'])
